# -*- coding: utf-8 -*-
#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

"""
Test the `ipalib.plugins.certprofile` module.
"""

import os

import pytest
import six

from ipalib import api, errors
from ipatests.util import prepare_config
from ipatests.test_xmlrpc.xmlrpc_test import XMLRPC_test, raises_exact
from ipatests.test_xmlrpc.tracker.certprofile_plugin import CertprofileTracker

if six.PY3:
    unicode = str

IPA_CERT_SUBJ_BASE = (
    api.Command.config_show()
    ['result']['ipacertificatesubjectbase'][0]
)

BASE_DIR = os.path.dirname(__file__)
CA_IPA_SERVICE_MODIFIED_TEMPLATE = os.path.join(
    BASE_DIR, 'data/caIPAserviceCert_mod.cfg.tmpl')

CA_IPA_SERVICE_MODIFIED_MALFORMED_TEMPLATE = os.path.join(
    BASE_DIR, 'data/caIPAserviceCert_mod_mal.cfg.tmpl')

CA_IPA_SERVICE_MALFORMED_TEMPLATE = os.path.join(
    BASE_DIR, 'data/caIPAserviceCert_mal.cfg.tmpl')

CA_IPA_SERVICE_XML_TEMPLATE = os.path.join(
    BASE_DIR, 'data/caIPAserviceCert.xml.tmpl')

RENAME_ERR_TEMPL = (
    u'certprofile {} cannot be deleted/modified: '
    'Certificate profiles cannot be renamed')


@pytest.fixture(scope='class')
def default_profile(request):
    name = 'caIPAserviceCert'
    desc = u'Standard profile for network services'
    tracker = CertprofileTracker(name, store=True, desc=desc)
    tracker.track_create()
    return tracker


@pytest.fixture(scope='class')
def user_profile(request):
    name = 'caIPAserviceCert_mod'
    profile_path = prepare_config(
        CA_IPA_SERVICE_MODIFIED_TEMPLATE,
        dict(
            ipadomain=api.env.domain,
            ipacertbase=IPA_CERT_SUBJ_BASE))

    tracker = CertprofileTracker(
        name, store=True, desc=u'Storing copy of a profile',
        profile=profile_path
    )

    return tracker.make_fixture(request)


@pytest.fixture(scope='class')
def malformed(request):
    name = u'caIPAserviceCert_mal'
    profile_path = prepare_config(
        CA_IPA_SERVICE_MALFORMED_TEMPLATE,
        dict(
            ipadomain=api.env.domain,
            ipacertbase=IPA_CERT_SUBJ_BASE))

    tracker = CertprofileTracker(name, store=True, desc=u'malformed profile',
                                 profile=profile_path)

    # Do not return with finalizer. There should be nothing to delete
    return tracker


@pytest.fixture(scope='class')
def xmlprofile(request):
    name = u'caIPAserviceCert_xml'
    profile_path = prepare_config(
        CA_IPA_SERVICE_XML_TEMPLATE,
        dict(
            ipadomain=api.env.domain,
            ipacertbase=IPA_CERT_SUBJ_BASE))

    tracker = CertprofileTracker(name, store=True, desc=u'xml format profile',
                                 profile=profile_path)

    return tracker


@pytest.mark.tier0
class TestDefaultProfile(XMLRPC_test):
    def test_default_profile_present(self, default_profile):
        default_profile.retrieve()

    def test_deleting_default_profile(self, default_profile):
        with pytest.raises(errors.ValidationError):
            default_profile.delete()

    def test_try_rename_by_setattr(self, default_profile):
        command = default_profile.make_update_command(
            updates=dict(setattr=u'cn=bogus'))
        errmsg = RENAME_ERR_TEMPL.format(default_profile.name)

        with raises_exact(errors.ProtectedEntryError(message=errmsg)):
            command()

    def test_try_rename_by_rename_option(self, default_profile):
        command = default_profile.make_update_command(dict(rename=u'bogus_id'))
        with pytest.raises(errors.OptionError):
            command()


@pytest.mark.tier1
class TestProfileCRUD(XMLRPC_test):
    def test_create_duplicate(self, user_profile):
        msg = u'Certificate Profile with name "{}" already exists'
        user_profile.ensure_exists()
        command = user_profile.make_create_command()
        with raises_exact(errors.DuplicateEntry(
                message=msg.format(user_profile.name))):
            command()

    def test_retrieve_simple(self, user_profile):
        user_profile.retrieve()

    def test_retrieve_all(self, user_profile):
        user_profile.retrieve(all=True)

    def test_export_profile(self, tmpdir, user_profile):
        profile = tmpdir.join('{}.cfg'.format(user_profile.name))

        command = user_profile.make_retrieve_command(out=unicode(profile))
        command()

        content = profile.read()
        assert user_profile.name in content

    def test_search_simple(self, user_profile):
        user_profile.find()

    def test_search_all(self, user_profile):
        user_profile.find(all=True)

    def test_update_store(self, user_profile):
        user_profile.update(
            dict(
                ipacertprofilestoreissued=False
            ),
            expected_updates=dict(
                ipacertprofilestoreissued=[u'FALSE']
            )
        )

    def test_update_description(self, user_profile):
        new_desc = u'new description'
        user_profile.update(
            dict(
                description=new_desc
            ),
            expected_updates=dict(
                description=[new_desc]
            )
        )

    def test_update_by_malformed_profile(self, user_profile):
        profile_path = prepare_config(
            CA_IPA_SERVICE_MODIFIED_MALFORMED_TEMPLATE,
            dict(
                ipadomain=api.env.domain,
                ipacertbase=IPA_CERT_SUBJ_BASE))

        with open(profile_path, ) as f:
            profile_content = f.read()
        command = user_profile.make_update_command(
            dict(file=unicode(profile_content)))

        with pytest.raises(errors.ExecutionError):
            command()

    def test_try_rename_by_setattr(self, user_profile):
        user_profile.ensure_exists()
        command = user_profile.make_update_command(
            updates=dict(setattr=u'cn=bogus'))
        errmsg = RENAME_ERR_TEMPL.format(user_profile.name)

        with raises_exact(errors.ProtectedEntryError(message=errmsg)):
            command()

    def test_delete(self, user_profile):
        user_profile.ensure_exists()
        user_profile.delete()

    def test_try_rename_by_rename_option(self, user_profile):
        user_profile.ensure_exists()
        command = user_profile.make_update_command(dict(rename=u'bogus_id'))
        with pytest.raises(errors.OptionError):
            command()


@pytest.mark.tier1
class TestMalformedProfile(XMLRPC_test):
    def test_malformed_import(self, malformed):
        with pytest.raises(errors.ExecutionError):
            malformed.create()


@pytest.mark.tier1
class TestImportFromXML(XMLRPC_test):
    def test_import_xml(self, xmlprofile):
        with pytest.raises(errors.ExecutionError):
            xmlprofile.ensure_exists()


# The initial user_profile configuration does not specify profileId.
# This is fine (it gets derived from the profile-id CLI argument),
# but this case was already tested in TestProfileCRUD.
#
# This test case tests various scenarios where the profileId *is*
# specified in the profile configuration.  These are:
#
# - mismatched profileId property (should fail)
# - multiple profileId properties (should fail)
# - one profileId property, matching given ID (should succeed)
#
@pytest.mark.tier1
class TestImportProfileIdHandling(XMLRPC_test):
    def test_import_with_mismatched_profile_id(self, user_profile):
        command = user_profile.make_create_command(
            extra_lines=['profileId=bogus']
        )
        with pytest.raises(errors.ValidationError):
            command()

    def test_import_with_multiple_profile_id(self, user_profile):
        # correct profile id, but two occurrences
        prop = u'profileId={}'.format(user_profile.name)
        command = user_profile.make_create_command(extra_lines=[prop, prop])
        with pytest.raises(errors.ValidationError):
            command()

    def test_import_with_correct_profile_id(self, user_profile):
        prop = u'profileId={}'.format(user_profile.name)
        command = user_profile.make_create_command(extra_lines=[prop])
        command()



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
RADIUS Proxy Servers

Manage RADIUS Proxy Servers.

IPA supports the use of an external RADIUS proxy server for krb5 OTP
authentications. This permits a great deal of flexibility when
integrating with third-party authentication services.

EXAMPLES:

 Add a new server:
   ipa radiusproxy-add MyRADIUS --server=radius.example.com:1812

 Find all servers whose entries include the string "example.com":
   ipa radiusproxy-find example.com

 Examine the configuration:
   ipa radiusproxy-show MyRADIUS

 Change the secret:
   ipa radiusproxy-mod MyRADIUS --secret

 Delete a configuration:
   ipa radiusproxy-del MyRADIUS
""")

register = Registry()


@register()
class radiusproxy(Object):
    takes_params = (
        parameters.Str(
            'cn',
            primary_key=True,
            label=_(u'RADIUS proxy server name'),
        ),
        parameters.Str(
            'description',
            required=False,
            label=_(u'Description'),
            doc=_(u'A description of this RADIUS proxy server'),
        ),
        parameters.Str(
            'ipatokenradiusserver',
            multivalue=True,
            label=_(u'Server'),
            doc=_(u'The hostname or IP (with or without port)'),
        ),
        parameters.Password(
            'ipatokenradiussecret',
            label=_(u'Secret'),
            doc=_(u'The secret used to encrypt data'),
        ),
        parameters.Int(
            'ipatokenradiustimeout',
            required=False,
            label=_(u'Timeout'),
            doc=_(u'The total timeout across all retries (in seconds)'),
        ),
        parameters.Int(
            'ipatokenradiusretries',
            required=False,
            label=_(u'Retries'),
            doc=_(u'The number of times to retry authentication'),
        ),
        parameters.Str(
            'ipatokenusermapattribute',
            required=False,
            label=_(u'User attribute'),
            doc=_(u'The username attribute on the user object'),
        ),
    )


@register()
class radiusproxy_add(Method):
    __doc__ = _("Add a new RADIUS proxy server.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'RADIUS proxy server name'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'A description of this RADIUS proxy server'),
        ),
        parameters.Str(
            'ipatokenradiusserver',
            multivalue=True,
            cli_name='server',
            label=_(u'Server'),
            doc=_(u'The hostname or IP (with or without port)'),
        ),
        parameters.Password(
            'ipatokenradiussecret',
            cli_name='secret',
            label=_(u'Secret'),
            doc=_(u'The secret used to encrypt data'),
            exclude=('cli', 'webui'),
            confirm=True,
        ),
        parameters.Int(
            'ipatokenradiustimeout',
            required=False,
            cli_name='timeout',
            label=_(u'Timeout'),
            doc=_(u'The total timeout across all retries (in seconds)'),
        ),
        parameters.Int(
            'ipatokenradiusretries',
            required=False,
            cli_name='retries',
            label=_(u'Retries'),
            doc=_(u'The number of times to retry authentication'),
        ),
        parameters.Str(
            'ipatokenusermapattribute',
            required=False,
            cli_name='userattr',
            label=_(u'User attribute'),
            doc=_(u'The username attribute on the user object'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class radiusproxy_del(Method):
    __doc__ = _("Delete a RADIUS proxy server.")

    takes_args = (
        parameters.Str(
            'cn',
            multivalue=True,
            cli_name='name',
            label=_(u'RADIUS proxy server name'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class radiusproxy_find(Method):
    __doc__ = _("Search for RADIUS proxy servers.")

    takes_args = (
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'cn',
            required=False,
            cli_name='name',
            label=_(u'RADIUS proxy server name'),
        ),
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'A description of this RADIUS proxy server'),
        ),
        parameters.Str(
            'ipatokenradiusserver',
            required=False,
            multivalue=True,
            cli_name='server',
            label=_(u'Server'),
            doc=_(u'The hostname or IP (with or without port)'),
        ),
        parameters.Password(
            'ipatokenradiussecret',
            required=False,
            cli_name='secret',
            label=_(u'Secret'),
            doc=_(u'The secret used to encrypt data'),
            exclude=('cli', 'webui'),
            confirm=True,
        ),
        parameters.Int(
            'ipatokenradiustimeout',
            required=False,
            cli_name='timeout',
            label=_(u'Timeout'),
            doc=_(u'The total timeout across all retries (in seconds)'),
        ),
        parameters.Int(
            'ipatokenradiusretries',
            required=False,
            cli_name='retries',
            label=_(u'Retries'),
            doc=_(u'The number of times to retry authentication'),
        ),
        parameters.Str(
            'ipatokenusermapattribute',
            required=False,
            cli_name='userattr',
            label=_(u'User attribute'),
            doc=_(u'The username attribute on the user object'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("name")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class radiusproxy_mod(Method):
    __doc__ = _("Modify a RADIUS proxy server.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'RADIUS proxy server name'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'A description of this RADIUS proxy server'),
        ),
        parameters.Str(
            'ipatokenradiusserver',
            required=False,
            multivalue=True,
            cli_name='server',
            label=_(u'Server'),
            doc=_(u'The hostname or IP (with or without port)'),
        ),
        parameters.Password(
            'ipatokenradiussecret',
            required=False,
            cli_name='secret',
            label=_(u'Secret'),
            doc=_(u'The secret used to encrypt data'),
            exclude=('cli', 'webui'),
            confirm=True,
        ),
        parameters.Int(
            'ipatokenradiustimeout',
            required=False,
            cli_name='timeout',
            label=_(u'Timeout'),
            doc=_(u'The total timeout across all retries (in seconds)'),
        ),
        parameters.Int(
            'ipatokenradiusretries',
            required=False,
            cli_name='retries',
            label=_(u'Retries'),
            doc=_(u'The number of times to retry authentication'),
        ),
        parameters.Str(
            'ipatokenusermapattribute',
            required=False,
            cli_name='userattr',
            label=_(u'User attribute'),
            doc=_(u'The username attribute on the user object'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'rename',
            required=False,
            label=_(u'Rename'),
            doc=_(u'Rename the RADIUS proxy server object'),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class radiusproxy_show(Method):
    __doc__ = _("Display information about a RADIUS proxy server.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'RADIUS proxy server name'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )



import six
from ipalib import _
if six.PY3:
    unicode = str

# Both constants can be used as masks against trust direction
# because bi-directional has two lower bits set.
TRUST_ONEWAY = 1
TRUST_BIDIRECTIONAL = 3

# Trust join behavior
# External trust -- allow creating trust to a non-root domain in the forest
TRUST_JOIN_EXTERNAL = 1

# We don't want to import any of Samba Python code here just for constants
# Since these constants set in MS-ADTS, we can rely on their stability
LSA_TRUST_ATTRIBUTE_NON_TRANSITIVE = 0x00000001

_trust_direction_dict = {
        1: _('Trusting forest'),
        2: _('Trusted forest'),
        3: _('Two-way trust')
}

_trust_status_dict = {
        True: _('Established and verified'),
        False: _('Waiting for confirmation by remote side')
}

_trust_type_dict_unknown = _('Unknown')

# Trust type is a combination of ipanttrusttype and ipanttrustattributes
# We shift trust attributes by 3 bits to left so bit 0 becomes bit 3 and
# 2+(1 << 3) becomes 10.
_trust_type_dict = {
        1: _('Non-Active Directory domain'),
        2: _('Active Directory domain'),
        3: _('RFC4120-compliant Kerberos realm'),
        10: _('Non-transitive external trust to a domain in '
              'another Active Directory forest'),
        11: _('Non-transitive external trust to an RFC4120-'
              'compliant Kerberos realm')
}


def trust_type_string(level, attrs):
    """
    Returns a string representing a type of the trust.
    The original field is an enum:
      LSA_TRUST_TYPE_DOWNLEVEL  = 0x00000001,
      LSA_TRUST_TYPE_UPLEVEL    = 0x00000002,
      LSA_TRUST_TYPE_MIT        = 0x00000003
    """
    transitive = int(attrs) & LSA_TRUST_ATTRIBUTE_NON_TRANSITIVE
    string = _trust_type_dict.get(int(level) | (transitive << 3),
                                  _trust_type_dict_unknown)
    return unicode(string)


def trust_direction_string(level):
    """
    Returns a string representing a direction of the trust.
    The original field is a bitmask taking two bits in use
      LSA_TRUST_DIRECTION_INBOUND  = 0x00000001,
      LSA_TRUST_DIRECTION_OUTBOUND = 0x00000002
    """
    string = _trust_direction_dict.get(int(level), _trust_type_dict_unknown)
    return unicode(string)


def trust_status_string(level):
    string = _trust_status_dict.get(level, _trust_type_dict_unknown)
    return unicode(string)



# Authors:
#   Pavel Zuna <pzuna@redhat.com>
#
# Copyright (C) 2009  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ipaclient.frontend import MethodOverride
from ipalib.plugable import Registry

register = Registry()


#@register()
class hbacrule_add_accesstime(MethodOverride):
    def output_for_cli(self, textui, result, cn, **options):
        textui.print_name(self.name)
        textui.print_dashed(
            'Added access time "%s" to HBAC rule "%s"' % (
                options['accesstime'], cn
            )
        )


#@register()
class hbacrule_remove_accesstime(MethodOverride):
    def output_for_cli(self, textui, result, cn, **options):
        textui.print_name(self.name)
        textui.print_dashed(
            'Removed access time "%s" from HBAC rule "%s"' % (
                options['accesstime'], cn
            )
        )



from __future__ import absolute_import

import os

import pytest

from ipapython.certdb import NSSDatabase, TRUSTED_PEER_TRUST_FLAGS
from ipaplatform.osinfo import osinfo

CERTNICK = 'testcert'

if osinfo.id == 'fedora':
    if osinfo.version_number >= (28,):
        NSS_DEFAULT = 'sql'
    else:
        NSS_DEFAULT = 'dbm'
else:
    NSS_DEFAULT = None


def create_selfsigned(nssdb):
    # create self-signed cert + key
    noisefile = os.path.join(nssdb.secdir, 'noise')
    with open(noisefile, 'wb') as f:
        f.write(os.urandom(64))
    try:
        nssdb.run_certutil([
            '-S', '-x',
            '-z', noisefile,
            '-k', 'rsa', '-g', '2048', '-Z', 'SHA256',
            '-t', 'CTu,Cu,Cu',
            '-s', 'CN=testcert',
            '-n', CERTNICK,
            '-m', '365',
        ])
    finally:
        os.unlink(noisefile)


def test_dbm_tmp():
    with NSSDatabase(dbtype='dbm') as nssdb:
        assert nssdb.dbtype == 'dbm'

        for filename in nssdb.filenames:
            assert not os.path.isfile(filename)
        assert not nssdb.exists()

        nssdb.create_db()
        for filename in nssdb.filenames:
            assert os.path.isfile(filename)
            assert os.path.dirname(filename) == nssdb.secdir
        assert nssdb.exists()

        assert os.path.basename(nssdb.certdb) == 'cert8.db'
        assert nssdb.certdb in nssdb.filenames
        assert os.path.basename(nssdb.keydb) == 'key3.db'
        assert os.path.basename(nssdb.secmod) == 'secmod.db'


def test_sql_tmp():
    with NSSDatabase(dbtype='sql') as nssdb:
        assert nssdb.dbtype == 'sql'

        for filename in nssdb.filenames:
            assert not os.path.isfile(filename)
        assert not nssdb.exists()

        nssdb.create_db()
        for filename in nssdb.filenames:
            assert os.path.isfile(filename)
            assert os.path.dirname(filename) == nssdb.secdir
        assert nssdb.exists()

        assert os.path.basename(nssdb.certdb) == 'cert9.db'
        assert nssdb.certdb in nssdb.filenames
        assert os.path.basename(nssdb.keydb) == 'key4.db'
        assert os.path.basename(nssdb.secmod) == 'pkcs11.txt'


def test_convert_db():
    with NSSDatabase(dbtype='dbm') as nssdb:
        assert nssdb.dbtype == 'dbm'

        nssdb.create_db()
        assert nssdb.exists()

        create_selfsigned(nssdb)

        oldcerts = nssdb.list_certs()
        assert len(oldcerts) == 1
        oldkeys = nssdb.list_keys()
        assert len(oldkeys) == 1

        nssdb.convert_db()
        assert nssdb.exists()

        assert nssdb.dbtype == 'sql'
        newcerts = nssdb.list_certs()
        assert len(newcerts) == 1
        assert newcerts == oldcerts
        newkeys = nssdb.list_keys()
        assert len(newkeys) == 1
        assert newkeys == oldkeys

        for filename in nssdb.filenames:
            assert os.path.isfile(filename)
            assert os.path.dirname(filename) == nssdb.secdir

        assert os.path.basename(nssdb.certdb) == 'cert9.db'
        assert nssdb.certdb in nssdb.filenames
        assert os.path.basename(nssdb.keydb) == 'key4.db'
        assert os.path.basename(nssdb.secmod) == 'pkcs11.txt'


def test_convert_db_nokey():
    with NSSDatabase(dbtype='dbm') as nssdb:
        assert nssdb.dbtype == 'dbm'
        nssdb.create_db()

        create_selfsigned(nssdb)

        assert len(nssdb.list_certs()) == 1
        assert len(nssdb.list_keys()) == 1
        # remove key, readd cert
        cert = nssdb.get_cert(CERTNICK)
        nssdb.run_certutil(['-F', '-n', CERTNICK])
        nssdb.add_cert(cert, CERTNICK, TRUSTED_PEER_TRUST_FLAGS)
        assert len(nssdb.list_keys()) == 0
        oldcerts = nssdb.list_certs()
        assert len(oldcerts) == 1

        nssdb.convert_db()
        assert nssdb.dbtype == 'sql'
        newcerts = nssdb.list_certs()
        assert len(newcerts) == 1
        assert newcerts == oldcerts
        assert nssdb.get_cert(CERTNICK) == cert
        newkeys = nssdb.list_keys()
        assert newkeys == ()

        for filename in nssdb.filenames:
            assert os.path.isfile(filename)
            assert os.path.dirname(filename) == nssdb.secdir

        old = os.path.join(nssdb.secdir, 'cert8.db')
        assert not os.path.isfile(old)
        assert os.path.isfile(old + '.migrated')

        assert os.path.basename(nssdb.certdb) == 'cert9.db'
        assert nssdb.certdb in nssdb.filenames
        assert os.path.basename(nssdb.keydb) == 'key4.db'
        assert os.path.basename(nssdb.secmod) == 'pkcs11.txt'


def test_auto_db():
    with NSSDatabase() as nssdb:
        assert nssdb.dbtype == 'auto'
        assert nssdb.filenames is None
        assert not nssdb.exists()
        with pytest.raises(RuntimeError):
            nssdb.list_certs()

        nssdb.create_db()
        assert nssdb.dbtype in ('dbm', 'sql')
        if NSS_DEFAULT is not None:
            assert nssdb.dbtype == NSS_DEFAULT
        assert nssdb.filenames is not None
        assert nssdb.exists()
        nssdb.list_certs()



# Authors:
#   Petr Vobornik <pvoborni@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
DNS tests
"""

from ipatests.test_webui.ui_driver import UI_driver
from ipatests.test_webui.ui_driver import screenshot
from ipatests.test_webui.data_dns import (
    ZONE_ENTITY, FORWARD_ZONE_ENTITY, CONFIG_ENTITY,
    ZONE_DEFAULT_FACET, ZONE_PKEY, ZONE_DATA, FORWARD_ZONE_PKEY,
    FORWARD_ZONE_DATA, RECORD_PKEY, A_IP, RECORD_ADD_DATA, RECORD_MOD_DATA,
    CONFIG_MOD_DATA
)
import pytest


@pytest.mark.tier1
class test_dns(UI_driver):

    def setup(self, *args, **kwargs):
        super(test_dns, self).setup(*args, **kwargs)

        if not self.has_dns():
            self.skip('DNS not configured')

    @screenshot
    def test_zone_record_crud(self):
        """
        Basic CRUD: dns
        """
        self.init_app()

        # add and mod zone
        self.basic_crud(ZONE_ENTITY, ZONE_DATA,
                        default_facet=ZONE_DEFAULT_FACET, delete=False)

        # add and mod record
        self.navigate_to_record(ZONE_PKEY)
        self.add_record(ZONE_ENTITY, RECORD_ADD_DATA,
                        facet=ZONE_DEFAULT_FACET, navigate=False)
        self.navigate_to_record(RECORD_PKEY)
        self.add_table_record('arecord', RECORD_MOD_DATA)

        # del record, del zone
        self.navigate_by_breadcrumb(ZONE_PKEY)
        self.delete_record(RECORD_PKEY)
        self.navigate_by_breadcrumb("DNS Zones")
        self.delete_record(ZONE_PKEY)

    @screenshot
    def test_forward_zone(self):
        """
        Forward DNS zones
        """
        self.init_app()

        # add and mod zone
        self.basic_crud(FORWARD_ZONE_ENTITY, FORWARD_ZONE_DATA, delete=False)

        # enable/disable
        self.navigate_to_record(FORWARD_ZONE_PKEY)

        self.disable_action()
        self.enable_action()
        self.action_list_action('add_permission')
        self.action_list_action('remove_permission')

        # del zone
        self.navigate_by_breadcrumb("DNS Forward Zones")
        self.delete_record(FORWARD_ZONE_PKEY)

    @screenshot
    def test_last_entry_deletion(self):
        """
        Test last entry deletion
        """
        self.init_app()
        self.add_record(ZONE_ENTITY, ZONE_DATA)
        self.navigate_to_record(ZONE_PKEY)
        self.add_record(ZONE_ENTITY, RECORD_ADD_DATA,
                        facet=ZONE_DEFAULT_FACET)
        self.navigate_to_record(RECORD_PKEY)
        self.delete_record(A_IP, parent=self.get_facet(), table_name='arecord')
        self.assert_dialog('message_dialog')
        self.dialog_button_click('ok')
        self.wait_for_request(n=2)
        self.assert_facet(ZONE_ENTITY, ZONE_DEFAULT_FACET)
        self.navigate_by_breadcrumb("DNS Zones")
        self.delete_record(ZONE_PKEY)

    @screenshot
    def test_config_crud(self):
        """
        Basic CRUD: dnsconfig
        """
        self.init_app()
        self.navigate_by_menu('network_services/dns/dnsconfig')
        self.mod_record(CONFIG_ENTITY, CONFIG_MOD_DATA)



# Authors:
#   Pavel Vomacka <pvomacka@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Vault tests
"""

from ipatests.test_webui.ui_driver import UI_driver
from ipatests.test_webui.ui_driver import screenshot
import ipatests.test_webui.data_vault as vault
import ipatests.test_webui.data_user as user
import ipatests.test_webui.data_group as group
import pytest


@pytest.mark.tier1
class vault_tasks(UI_driver):

    def prep_service_data(self):

        host = self.config.get('ipa_server')
        realm = self.config.get('ipa_realm')
        pkey = 'itest'

        return {
            'entity': 'service',
            'pkey': '%s/%s@%s' % (pkey, host, realm),
            'add': [
                ('textbox', 'service', pkey),
                ('combobox', 'host', host)
            ]
        }

    def prepare_vault_service_data(self, data):
        s_data = self.prep_service_data()
        service = s_data['pkey']

        serv_field = [('combobox', 'service', service)]

        data['add'].extend(serv_field)

    def prepare_vault_user_data(self, data, user='admin'):
        user_field = [('combobox', 'username', user)]

        data['add'].extend(user_field)


@pytest.mark.tier1
class test_vault(vault_tasks):

    def setup(self, *args, **kwargs):
        super(test_vault, self).setup(*args, **kwargs)
        if not self.has_kra():
            self.skip('KRA not configured')

    @screenshot
    def test_crud(self):
        """
        Basic basic CRUD: user vault
        """
        self.init_app()
        self.prepare_vault_user_data(vault.DATA)
        self.basic_crud(vault.ENTITY, vault.DATA)

    @screenshot
    def test_add_service_vault(self):
        """
        Add Service vault
        """
        self.init_app()

        # Add itest service
        s_data = self.prep_service_data()
        self.add_record(s_data['entity'], s_data)

        self.prepare_vault_service_data(vault.DATA2)

        # Add and remove service vault
        self.add_record(vault.ENTITY, vault.DATA2, facet=vault.DATA2['facet'],
                        delete=True)

        # Remove test service
        self.navigate_to_entity(s_data['entity'])
        self.delete_record(s_data['pkey'])

    @screenshot
    def test_add_shared_vault(self):
        """
        Add Shared vault
        """
        self.init_app()

        # Add shared vault
        self.add_record(vault.ENTITY, vault.DATA3, facet=vault.DATA3['facet'],
                        delete=True)

    @screenshot
    def test_member_owner_vault(self):
        """
        Add User Vault and try to add member and owner
        """
        def fill_tables():
            self.add_table_associations('member_user', [user.PKEY])
            self.add_table_associations('member_group', [group.PKEY])
            self.add_table_associations('member_service', [s_data['pkey']])
            self.add_table_associations('owner_user', [user.PKEY])
            self.add_table_associations('owner_group', [group.PKEY])
            self.add_table_associations('owner_service', [s_data['pkey']])

        # Add user
        self.init_app()
        self.add_record(user.ENTITY, user.DATA)

        # Prepare items - user already exists
        s_data = self.prep_service_data()
        self.add_record(s_data['entity'], s_data)
        self.add_record(group.ENTITY, group.DATA)

        # USER
        # Add user vault
        self.add_record(vault.ENTITY, vault.DATA, facet='user_search')

        # Navigate to record
        self.navigate_to_record(vault.DATA['pkey'])

        # Try add values into table
        fill_tables()

        # Remove user vault record
        self.navigate_to_entity(vault.ENTITY, vault.DATA['facet'])
        self.delete_record(vault.PKEY)

        # SERVICE
        # Add service vault
        self.prepare_vault_service_data(vault.DATA2)
        self.add_record(vault.ENTITY, vault.DATA2, facet=vault.DATA2['facet'])

        # Navigate to record
        self.navigate_to_record(vault.DATA2['pkey'])

        # Try add values into table
        fill_tables()

        # Remove service vault record
        self.navigate_to_entity(vault.ENTITY, vault.DATA2['facet'])
        self.delete_record(vault.DATA2['pkey'])

        # SHARED
        # Add shared vault
        self.add_record(vault.ENTITY, vault.DATA3, facet=vault.DATA3['facet'])

        # Navigate to record
        self.navigate_to_record(vault.DATA3['pkey'])

        # Try add values into table
        fill_tables()

        # Remove shared vault record
        self.navigate_to_entity(vault.ENTITY, vault.DATA3['facet'])
        self.delete_record(vault.DATA3['pkey'])

        # Clean up
        self.navigate_to_entity(s_data['entity'])
        self.delete_record(s_data['pkey'])
        self.navigate_to_entity(user.ENTITY)
        self.delete_record(user.PKEY)
        self.navigate_to_entity(group.ENTITY)
        self.delete_record(group.PKEY)




# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#

BAD_IP_MSG = "Not a valid IP address"
BAD_HOSTNAME_MSG = "only letters, numbers, '-' are allowed"
BAS_HOSTNAME_SPACE_MSG = "Leading and trailing spaces are not allowed"

empty_hostname = {
    'pkey': 'empty_hostname',
    'add': [
        ('textbox', 'hostname', ''),
    ],
}

empty_domain = {
    'pkey': 'empty_domain',
    'add': [
        ('textbox', 'hostname', 'itest-empty-domain'),
        ('textbox', 'dnszone', ''),
    ],
}

hostname_tilde = {
    'pkey': 'tilde_hostname',
    'add': [
        ('textbox', 'hostname', '~tilde'),
    ],
}

hostname_dash = {
    'pkey': 'dash_hostname',
    'add': [
        ('textbox', 'hostname', '-dash'),
    ],
}

hostname_leading_space = {
    'pkey': 'leading_space',
    'add': [
        ('textbox', 'hostname', ' leading_space'),
    ],
}

hostname_trailing_space = {
    'pkey': 'trailing_space',
    'add': [
        ('textbox', 'hostname', 'trailing_space '),
    ],
}

ip_alpha = {
    'pkey': 'ip_alpha',
    'add': [
        ('textbox', 'hostname', 'ip-field-test'),
        ('textbox', 'ip_address', 'abc.10.12.14'),
    ],
}

ip_many_oct = {
    'pkey': 'ip_many',
    'add': [
        ('textbox', 'hostname', 'ip-field-test'),
        ('textbox', 'ip_address', '10.10.10.1.10'),
    ],
}

ip_bad_oct = {
    'pkey': 'ip_bad_octal',
    'add': [
        ('textbox', 'hostname', 'ip-field-test'),
        ('textbox', 'ip_address', '10.0.378.1'),
    ],
}

ip_special_char = {
    'pkey': 'ip_special',
    'add': [
        ('textbox', 'hostname', 'ip-field-test'),
        ('textbox', 'ip_address', '10.0.##.1'),
    ],
}

mod_desc = [
    ('textarea', 'description', 'description in details'),
]

mod_desc_m = [
    ('textarea', 'description', 'description never appear'),
]

mod_locality = [
    ('textbox', 'l', 'Brno Office'),
]

mod_location = [
    ('textbox', 'nshostlocation', 'Brno Office'),
]

mod_platform = [
    ('textbox', 'nshardwareplatform', 'x86_64'),
]

mod_os = [
    ('textbox', 'nsosversion', 'FEDORA RHEL 277'),
]

otp_alpha = [
    ('password', 'userpassword', 'alpha'),
    ('password', 'password2', 'alpha'),
]

otp_num = [
    ('password', 'userpassword', '1234'),
    ('password', 'password2', '1234'),
]

otp_alphanum = [
    ('password', 'userpassword', 'abc123'),
    ('password', 'password2', 'abc123'),
]

otp_special = [
    ('password', 'userpassword', '@#$'),
    ('password', 'password2', '@#$'),
]

otp_mixed = [
    ('password', 'userpassword', 'AbC12D'),
    ('password', 'password2', 'AbC12D'),
]

ssh_nomod_error = "no modifications to be performed"
ssh_invalid_error = "invalid SSH public key"
ssh_rsa = 'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQCnAOLat' + \
          'ncsaDxF+ldDhjNdPRDWXKsiZUz6Y49LjPnEr9p4Th24dZ' + \
          '7ZuvOVjhXDSkivh6MRunWZC+MXxRo1lDZgkCSyQfkMq0E' + \
          'u6xkubPg3tYAdrFBZIcIl5CUNerqYMdTz2hyTq6HAR/qs' + \
          '6oRbtzUemwHLPo3duqDRLWQoojP+tI8I2IEXnOO2N5oxq' + \
          'YGWAUe7bGXS/O2ukGfclt8/BfVw9e6eqHqlc7tKGqEctn' + \
          'imlsbG291ctNgco8FsvCnV5EOti/O0rLdkTmm66j7WCFj' + \
          'D9gJncfkAzxc+itWE4eUg/0B5ICIeRrFl5obD8Vu3LzTQ' + \
          '4yKiwaUnY5ngXgWBoFq9 root@sshkey.testipa.ipa'
ssh_dsa = 'ssh-dss AAAAB3NzaC1kc3MAAACBAO3OqwC1eNedXVJ57' + \
          '/a+Q/BfVcbZiJcTxhVP6TnIIQXmI+YSu685gLXEHWEAX1' + \
          '9+8eQuvUSmgWViuskErCXliE1c4PVyrwkf/2UMsH+hFaj' + \
          '0jlAM4APzizSvHC59hjpr5ktPyrv1arBXGYRuWyZNphJZ' + \
          'OFbqK2DHZbz1jvhD4uz5AAAAFQDasuSv8Dkn2Khqek0U3' + \
          'EAHUaUL2wAAAIB5Wr4r7z4ZyaSoaxfiUvvKg49FCeGjrY' + \
          'jRbYN/PazAn/X0rPcGqpaF3u5FmxXP7vhvlvECZvveK7T' + \
          'FIJVz1DSKHMRu8886akKLegF1zhhjrnjN7Q4vHbwkhsCI' + \
          'aV+4rlJa7B32girkSltlooP/qWMnRde0aJIf20Zhq/IF9' + \
          'oj49AAAAIBsKrdE+nxubD13+BdX07Sq6wAPVa9RVCISqE' + \
          'simlCvopStg8vNuNfGi9swmyFyNjSMiZEgoxH2cLRME4+' + \
          'xzn7THVrmE6OQ/Duz/mQAnDvt1N0Qw4jNxv0WqoT0kz7X' + \
          '21L5Dmg5qy4qdEvlcOkVI9gMrIrXhwGb+Vj8XEGtWcNmJ' + \
          'w== root@sshkey.testipa.ipa'
ssh_empty = ''
ssh_invalid = 'ff99cc1234dec invalid'

csr_invalid_error = 'Base64 decoding failed'
csr_invalid = 'invalid cert'
csr_other_host_error = 'Base64 decoding failed'
csr_other_host = '''-----BEGIN NEW CERTIFICATE REQUEST-----
MIICpjCCAY4CAQAwMDEQMA4GA1UEChMHSVBBLklQQTEcMBoGA1UEAxMTYW5vdGhl
cmhvc3QuaXBhLmlwYTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAM5Z
xiBOxo0W107maJt84m4BrkOFErCi0Mk4UQABMAfg/Sbj/+nYL19CA/IgSy4NoCnt
0RK1IZcFvSzHNhJmwpyRcmaOIbEsjes35rYYA2LKV3QVBp14284tJN5xRHztuL9B
0NDaSuZOG4JERHJl7JBGOzs4mj3FkI+Ci92d/zi+vpI+T0b26BGejcpU98zkVKxE
ktXNHqZp/QV7EHsqaZDdIPGTORZokZnU3VFsbUnCLDyghg3+75t+Wq4sJvwL1Y9j
btO5cQLWTJiOotk1Ies2A5nrp89CpMP45ERtZmoe+G3WeWgW9Nqr182kF5NjlC/O
sHKz4bP9hT9Z6bk4J3ECAwEAAaAxMC8GCSqGSIb3DQEJDjEiMCAwHgYDVR0RBBcw
FYITYW5vdGhlcmhvc3QuaXBhLmlwYTANBgkqhkiG9w0BAQsFAAOCAQEAoEQNqnts
Ob5fTPZRQQo8ygoKa+4GXMjM/Ue2SYs2zOa1/aYeI6JVzWzWH9xHFNvhOkdhu154
9fefKPtFKeyRTRz60KjSGcHyawDmoWyVYMPgFwmWp1lceFDEy0SlCnB58iXuxYEU
mwlXmODQR1hQxLuo5Ow3Hy0Djyml7gh7DA/iHP7WrOJH3PwTegxAFFixIj7K6DYK
3Kaeng72Ht8vQeTEh0Fq4rcfIdlW6tjWywLqLqCjtwhNkak4tJna6M9/3yjeyEnk
/w7Ya8CyOwlTaCvN8cjnBTxXWVVh+lIaPhujxG4UVtOMqaI30EkMIMHrocCUNnRd
2e8CMvHPLREqJw==
-----END NEW CERTIFICATE REQUEST-----'''

krb_enrolled = 'Kerberos Key Present, Host Provisioned'
krb_not_enrolled = 'Kerberos Key Not Present'



# Authors:
#   Petr Vobornik <pvoborni@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Selfservice tests
"""

from ipatests.test_webui.ui_driver import UI_driver
from ipatests.test_webui.ui_driver import screenshot
from ipatests.test_webui import data_selfservice
import ipatests.test_webui.data_user as user
import pytest

try:
    from selenium.webdriver.common.by import By
    from selenium.webdriver.common.keys import Keys
    from selenium.webdriver.common.action_chains import ActionChains
except ImportError:
    pass

ENTRY_EXIST = 'This entry already exists'
FIELD_REQ = 'Required field'
INV_NAME = ("invalid 'name': Leading and trailing spaces are "
            "not allowed")
ERR_INCLUDE = 'May only contain letters, numbers, -, _, and space'
SERVICE_ADDED = 'Self Service Permission successfully added'


def reset_passwd(self, login, pwd):
            self.navigate_to_entity(user.ENTITY)
            self.navigate_to_record(login)
            self.action_list_action('reset_password', False)
            self.fill_password('password', pwd)
            self.fill_password('password2', pwd)
            self.dialog_button_click('confirm')


@pytest.mark.tier1
class test_selfservice(UI_driver):

    @screenshot
    def test_crud(self):
        """
        Basic CRUD: selfservice entity
        """
        self.init_app()
        self.basic_crud(data_selfservice.ENTITY, data_selfservice.DATA)

    @screenshot
    def test_add_all_attr(self):
        """
        Add self service with all attribute
        """
        self.init_app()
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA_ALL,
                        delete=True)

    @screenshot
    def test_add_scenarios(self):
        """
        Test various add scenarios
        """

        # Add self service with Add and Add Another button
        self.init_app()
        records = [data_selfservice.DATA, data_selfservice.DATA1]
        self.add_record(data_selfservice.ENTITY, records)

        # check if record is added
        for record in records:
            self.assert_record(record['pkey'])

        # cleanup
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.select_multiple_records(records)
        self.facet_button_click('remove')
        self.dialog_button_click('ok')
        for record in records:
            self.assert_record(record['pkey'], negative=True)

        # Add self service with Add and edit button
        self.add_record(data_selfservice.ENTITY,
                        data_selfservice.DATA1,
                        dialog_btn='add_and_edit')

        # cleanup
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.delete(data_selfservice.ENTITY, [data_selfservice.DATA1])

        # Add self service with Add and cancel button
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA1,
                        dialog_btn='cancel')

    @screenshot
    def test_undo_reset(self):
        """
        Test undo and reset operation
        """
        # Add self service permission and perform undo
        self.init_app()
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA1)
        self.navigate_to_record(data_selfservice.PKEY1)
        self.fill_fields(data_selfservice.DATA2)
        self.click_undo_button('attrs')
        self.wait_for_request()
        # if undo succeed, 'save' button remains disabled
        self.assert_facet_button_enabled('save', enabled=False)

        # Add self service permission and perform reset
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA1)
        self.navigate_to_record(data_selfservice.PKEY1)
        self.fill_fields(data_selfservice.DATA2)
        self.facet_button_click('revert')
        # if revert succeed, 'save' button remains disabled
        self.assert_facet_button_enabled('save', enabled=False)

    @screenshot
    def test_permission_negative(self):
        """
        Negative test cases for self service permission
        """
        self.init_app()

        # try to add duplicate entry
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA1)
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA1,
                        negative=True, pre_delete=False)
        self.assert_last_error_dialog(ENTRY_EXIST)
        self.close_all_dialogs()
        self.delete(data_selfservice.ENTITY, [data_selfservice.DATA1])
        self.close_notifications()

        # try to add permission without name and attribute
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.wait_for_request()
        self.facet_button_click('add')
        self.dialog_button_click('add')
        self.assert_field_validation(FIELD_REQ, field='aciname')
        self.dialog_button_click('cancel')

        # try to add permission without name but having attribute
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.wait_for_request()
        self.facet_button_click('add')
        self.check_option('attrs', 'displayname')
        self.dialog_button_click('add')
        self.assert_field_validation(FIELD_REQ, field='aciname')
        self.dialog_button_click('cancel')

        # try to add permission without having attribute.
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.wait_for_request()
        self.facet_button_click('add')
        self.fill_textbox('aciname', data_selfservice.DATA1['pkey'])
        self.dialog_button_click('add')
        self.assert_field_validation(FIELD_REQ, field='attrs')
        self.dialog_button_click('cancel')

        # try to add aciname with leading space.
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.wait_for_request()
        self.facet_button_click('add')
        self.fill_textbox('aciname',
                          ' {}'.format(data_selfservice.DATA1['pkey']))
        self.check_option('attrs', 'audio')
        self.dialog_button_click('add')
        self.assert_last_error_dialog(INV_NAME)
        self.close_all_dialogs()

        # try to add aciname with trailing space
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.wait_for_request()
        self.facet_button_click('add')
        self.fill_textbox('aciname',
                          '{} '.format(data_selfservice.DATA1['pkey']))
        self.check_option('attrs', 'audio')
        self.dialog_button_click('add')
        self.assert_last_error_dialog(INV_NAME)
        self.dialog_button_click('cancel')
        self.dialog_button_click('cancel')

        # try to add aciname with special char
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.wait_for_request()
        self.facet_button_click('add')
        self.fill_textbox('aciname', '#%^')
        self.dialog_button_click('add')
        self.assert_field_validation(ERR_INCLUDE, field='aciname')
        self.dialog_button_click('cancel')

        # try to modify pesmission by removing all attributes
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA1)
        self.navigate_to_record(data_selfservice.DATA1['pkey'])
        self.check_option('attrs', 'businesscategory')
        self.facet_button_click('save')
        self.assert_field_validation(FIELD_REQ, field='attrs')

    @screenshot
    def test_del_multiple_permission(self):
        """
        Try to delete multiple self service permission
        """
        self.init_app()
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA1)
        self.add_record(data_selfservice.ENTITY, data_selfservice.DATA)
        self.delete(data_selfservice.ENTITY,
                    [data_selfservice.DATA1,
                     data_selfservice.DATA])
        # check if record deleted
        for key in [data_selfservice.DATA1['pkey'],
                    data_selfservice.DATA['pkey']]:
            self.assert_record(key, negative=True)

    @screenshot
    def test_permission_using_enter_key(self):
        """
        Try to add/delete persmission using enter key
        """
        # try to add using enter key
        self.init_app()
        self.add_record(data_selfservice.ENTITY,
                        data_selfservice.DATA1,
                        dialog_btn=None)
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ENTER).perform()
        self.wait()
        self.assert_notification(assert_text=SERVICE_ADDED)
        self.assert_record(data_selfservice.DATA1['pkey'])
        self.close_notifications()

        # try to delete using enter key
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.select_record(data_selfservice.DATA1['pkey'])
        self.facet_button_click('remove')
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ENTER).perform()
        self.wait()
        self.assert_notification(assert_text='1 item(s) deleted')
        self.close_notifications()

    @screenshot
    def test_reset_sshkey_permsission(self):
        """
        Try to delete sshkey after altering sshkey permission
        """
        pwd = self.config.get('ipa_password')

        self.init_app()
        self.add_record(user.ENTITY, user.DATA, navigate=False)
        reset_passwd(self, user.PKEY, pwd)
        self.logout()
        self.login(user.PKEY, password=pwd, new_password=pwd)
        self.add_sshkey_to_record(user.SSH_RSA, user.PKEY, navigate=True)
        self.assert_num_ssh_keys(1)
        self.close_notifications()
        self.logout()

        self.init_app()
        self.navigate_to_entity(data_selfservice.ENTITY)
        self.navigate_to_record('Users can manage their own SSH public keys')
        # to pass the validator, check some option from the options
        self.check_option('attrs', 'carlicense')
        # uncheck the ipasshpubkey checkbox.
        self.check_option('attrs', 'ipasshpubkey')
        self.facet_button_click('save')
        self.close_notifications()
        self.logout()

        # check if delete button is visible for ssh key.
        self.login(user.PKEY, password=pwd)
        s = "div[name='ipasshpubkey'].multivalued-widget"
        facet = self.get_facet()
        btn = self.find(s, By.CSS_SELECTOR, facet, strict=True)
        assert btn.is_displayed is not True



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

from ipatests.test_xmlrpc import objectclasses
from ipatests.test_xmlrpc.xmlrpc_test import fuzzy_uuid

from ipatests.test_xmlrpc.tracker.base import Tracker
from ipatests.util import assert_deepequal

from ipalib import api
from ipapython.dn import DN


class HostGroupTracker(Tracker):
    """ Class for tracking hostgroups """
    retrieve_keys = {u'dn', u'cn', u'member_host', u'description',
                     u'member_hostgroup', u'memberindirect_host'}
    retrieve_all_keys = retrieve_keys | {u'ipauniqueid', u'objectclass',
                                         u'mepmanagedentry'}

    create_keys = retrieve_all_keys
    update_keys = retrieve_keys - {u'dn'}

    add_member_keys = retrieve_keys | {u'member_host'}

    find_keys = {
        u'dn', u'cn', u'description',
    }
    find_all_keys = {
        u'dn', u'cn', u'member_host', u'description', u'member_hostgroup',
        u'memberindirect_host', u'ipauniqueid', u'objectclass',
        u'mepmanagedentry',
    }

    def __init__(self, name, description=u'HostGroup desc'):
        super(HostGroupTracker, self).__init__(default_version=None)
        self.cn = name
        self.description = description
        self.dn = DN(('cn', self.cn), ('cn', 'hostgroups'),
                     ('cn', 'accounts'), api.env.basedn)

    def make_create_command(self,
                            force=True, *args, **kwargs):
        """ Make function that creates a hostgroup using 'hostgroup-add' """
        return self.make_command('hostgroup_add', self.cn,
                                 description=self.description,
                                 *args, **kwargs)

    def make_delete_command(self):
        """ Make function that deletes a hostgroup using 'hostgroup-del' """
        return self.make_command('hostgroup_del', self.cn)

    def make_retrieve_command(self, all=False, raw=False):
        """ Make function that retrieves a hostgroup using 'hostgroup-show' """
        return self.make_command('hostgroup_show', self.cn, all=all)

    def make_find_command(self, *args, **kwargs):
        """ Make function that searches for a hostgroup
            using 'hostgroup-find' """
        return self.make_command('hostgroup_find', *args, **kwargs)

    def make_update_command(self, updates):
        """ Make function that updates a hostgroup using 'hostgroup-mod' """
        return self.make_command('hostgroup_mod', self.cn, **updates)

    def make_add_member_command(self, options={}):
        """ Make function that adds a member to a hostgroup """
        return self.make_command('hostgroup_add_member', self.cn, **options)

    def make_remove_member_command(self, options={}):
        """ Make function that removes a member from a hostgroup """
        return self.make_command('hostgroup_remove_member', self.cn, **options)

    def track_create(self):
        """ Updates expected state for hostgroup creation"""
        self.attrs = dict(
            dn=self.dn,
            mepmanagedentry=[DN(('cn', self.cn), ('cn', 'ng'),
                                ('cn', 'alt'), api.env.basedn)],
            cn=[self.cn],
            description=[self.description],
            ipauniqueid=[fuzzy_uuid],
            objectclass=objectclasses.hostgroup,
            )
        self.exists = True

    def add_member(self, options):
        """ Add a member host to hostgroup and perform check """
        if u'host' in options:
            try:
                self.attrs[u'member_host'] =\
                    self.attrs[u'member_host'] + [options[u'host']]
            except KeyError:
                self.attrs[u'member_host'] = [options[u'host']]
            # search for hosts in the target hostgroup and
            # add them as memberindirect hosts
        elif u'hostgroup' in options:
            try:
                self.attrs[u'member_hostgroup'] =\
                    self.attrs[u'member_hostgroup'] + [options[u'hostgroup']]
            except KeyError:
                self.attrs[u'member_hostgroup'] = [options[u'hostgroup']]

        command = self.make_add_member_command(options)
        result = command()
        self.check_add_member(result)

    def remove_member(self, options):
        """ Remove a member host from hostgroup and perform check """
        if u'host' in options:
            self.attrs[u'member_host'].remove(options[u'host'])
        elif u'hostgroup' in options:
            self.attrs[u'member_hostgroup'].remove(options[u'hostgroup'])

        try:
            if not self.attrs[u'member_host']:
                del self.attrs[u'member_host']
        except KeyError:
            pass
        try:
            if not self.attrs[u'member_hostgroup']:
                del self.attrs[u'member_hostgroup']
        except KeyError:
            pass

        command = self.make_remove_member_command(options)
        result = command()
        self.check_remove_member(result)

    def update(self, updates, expected_updates=None):
        """Helper function to update this user and check the result

        Overriding Tracker method for setting self.attrs correctly;
         * most attributes stores its value in list
         * the rest can be overridden by expected_updates
         * allow deleting parametrs if update value is None
        """
        if expected_updates is None:
            expected_updates = {}

        self.ensure_exists()
        command = self.make_update_command(updates)
        result = command()

        for key, value in updates.items():
            if value is None:
                del self.attrs[key]
            else:
                self.attrs[key] = [value]
        for key, value in expected_updates.items():
            if value is None:
                del self.attrs[key]
            else:
                self.attrs[key] = value

        self.check_update(
            result,
            extra_keys=set(updates.keys()) | set(expected_updates.keys())
        )

    def check_create(self, result):
        """ Checks 'hostgroup_add' command result """
        assert_deepequal(dict(
            value=self.cn,
            summary=u'Added hostgroup "%s"' % self.cn,
            result=self.filter_attrs(self.create_keys)
            ), result)

    def check_delete(self, result):
        """ Checks 'hostgroup_del' command result """
        assert_deepequal(dict(
            value=[self.cn],
            summary=u'Deleted hostgroup "%s"' % self.cn,
            result=dict(failed=[]),
            ), result)

    def check_retrieve(self, result, all=False, raw=False):
        """ Checks 'hostgroup_show' command result """
        if all:
            expected = self.filter_attrs(self.retrieve_all_keys)
        else:
            expected = self.filter_attrs(self.retrieve_keys)

        assert_deepequal(dict(
            value=self.cn,
            summary=None,
            result=expected
            ), result)

    def check_find(self, result, all=False, raw=False):
        """ Checks 'hostgroup_find' command result """
        if all:
            expected = self.filter_attrs(self.find_all_keys)
        else:
            expected = self.filter_attrs(self.find_keys)

        assert_deepequal(dict(
            count=1,
            truncated=False,
            summary=u'1 hostgroup matched',
            result=[expected],
        ), result)

    def check_update(self, result, extra_keys={}):
        """ Checks 'hostgroup_mod' command result """
        assert_deepequal(dict(
            value=self.cn,
            summary=u'Modified hostgroup "%s"' % self.cn,
            result=self.filter_attrs(self.update_keys | set(extra_keys))
        ), result)

    def check_add_member(self, result):
        """ Checks 'hostgroup_add_member' command result """
        assert_deepequal(dict(
            completed=1,
            failed={u'member': {u'host': (), u'hostgroup': ()}},
            result=self.filter_attrs(self.add_member_keys)
        ), result)

    def check_add_member_negative(self, result, options):
        """ Checks 'hostgroup_add_member' command result
        when expected result is failure of the operation"""
        expected = dict(
            completed=0,
            failed={u'member': {u'hostgroup': (), u'user': ()}},
            result=self.filter_attrs(self.add_member_keys)
        )
        if u'host' in options:
            expected[u'failed'][u'member'][u'host'] = [(
                options[u'host'], u'no such entry')]
        elif u'hostgroup' in options:
            expected[u'failed'][u'member'][u'hostgroup'] = [(
                options[u'hostgroup'], u'no such entry')]

        assert_deepequal(expected, result)

    def check_remove_member_negative(self, result, options):
        """ Checks 'hostgroup_remove_member' command result
        when expected result is failure of the operation"""
        expected = dict(
            completed=0,
            failed={u'member': {u'hostgroup': (), u'host': ()}},
            result=self.filter_attrs(self.add_member_keys)
        )
        if u'user' in options:
            expected[u'failed'][u'member'][u'host'] = [(
                options[u'user'], u'This entry is not a member')]
        elif u'hostgroup' in options:
            expected[u'failed'][u'member'][u'hostgroup'] = [(
                options[u'hostgroup'], u'This entry is not a member')]

        assert_deepequal(expected, result)

    def check_remove_member(self, result):
        """ Checks 'hostgroup_remove_member' command result """
        self.check_add_member(result)



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

from __future__ import absolute_import

import logging
import os
import tempfile

from ipalib import Registry
from ipalib import Updater
from ipalib.install import certmonger
from ipaplatform.paths import paths
from ipapython.certdb import NSSDatabase
from ipaserver.install import cainstance

logger = logging.getLogger(__name__)

register = Registry()


@register()
class update_ra_cert_store(Updater):
    """
    Moves the ipaCert store from /etc/httpd/alias RA_AGENT_PEM, RA_AGENT_KEY
    files
    """

    def execute(self, **options):
        ra_nick = 'ipaCert'
        ca_enabled = self.api.Command.ca_is_enabled()['result']
        if not ca_enabled:
            return False, []

        certdb = NSSDatabase(nssdir=paths.HTTPD_ALIAS_DIR)
        if not certdb.has_nickname(ra_nick):
            # Nothign to do
            return False, []
        elif os.path.exists(paths.RA_AGENT_PEM):
            # even though the certificate file exists, we will overwrite it
            # as it's probabably something wrong anyway
            logger.warning(
                "A certificate with the nickname 'ipaCert' exists in "
                "the old '%s' NSS database as well as in the new "
                "PEM file '%s'",
                paths.HTTPD_ALIAS_DIR, paths.RA_AGENT_PEM)

        _fd, p12file = tempfile.mkstemp(dir=certdb.secdir)
        # no password is necessary as we will be saving it in clear anyway
        certdb.export_pkcs12(ra_nick, p12file, pkcs12_passwd='')

        # stop tracking the old cert and remove it
        certmonger.stop_tracking(paths.HTTPD_ALIAS_DIR, nickname=ra_nick)
        certdb.delete_cert(ra_nick)
        if os.path.exists(paths.OLD_KRA_AGENT_PEM):
            os.remove(paths.OLD_KRA_AGENT_PEM)

        # get the private key and certificate from the file and start
        # tracking it in certmonger
        ca = cainstance.CAInstance()
        ca.import_ra_cert(p12file)

        os.remove(p12file)

        return False, []



#
# Copyright (C) 2017  FreeIPA Contributors see COPYING for license
#

"""
This module contains default Debian-specific implementations of system tasks.
"""

from __future__ import absolute_import

from ipaplatform.base.tasks import BaseTaskNamespace
from ipaplatform.redhat.tasks import RedHatTaskNamespace

from ipapython import ipautil

class DebianTaskNamespace(RedHatTaskNamespace):
    @staticmethod
    def restore_pre_ipa_client_configuration(fstore, statestore,
                                             was_sssd_installed,
                                             was_sssd_configured):
        try:
            ipautil.run(["pam-auth-update",
                         "--package", "--remove", "mkhomedir"])
        except ipautil.CalledProcessError:
            return False
        return True

    @staticmethod
    def set_nisdomain(nisdomain):
        # Debian doesn't use authconfig, nothing to set
        return True

    @staticmethod
    def modify_nsswitch_pam_stack(sssd, mkhomedir, statestore, sudo=True):
        if mkhomedir:
            try:
                ipautil.run(["pam-auth-update",
                             "--package", "--enable", "mkhomedir"])
            except ipautil.CalledProcessError:
                return False
            return True
        else:
            return True

    @staticmethod
    def modify_pam_to_use_krb5(statestore):
        # Debian doesn't use authconfig, this is handled by pam-auth-update
        return True

    @staticmethod
    def backup_auth_configuration(path):
        # Debian doesn't use authconfig, nothing to backup
        return True

    @staticmethod
    def restore_auth_configuration(path):
        # Debian doesn't use authconfig, nothing to restore
        return True

    def migrate_auth_configuration(self, statestore):
        # Debian doesn't have authselect
        return True

    @staticmethod
    def parse_ipa_version(version):
        return BaseTaskNamespace.parse_ipa_version(version)

    def configure_httpd_wsgi_conf(self):
        # Debian doesn't require special mod_wsgi configuration
        pass

    def setup_httpd_logging(self):
        # Debian handles httpd logging differently
        pass

    def configure_pkcs11_modules(self, fstore):
        # Debian doesn't use p11-kit
        pass

    def restore_pkcs11_modules(self, fstore):
        pass

tasks = DebianTaskNamespace()



# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2008  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Defines the expected objectclass for various entries.
"""

user_base = [
    u'top',
    u'person',
    u'organizationalperson',
    u'inetorgperson',
    u'inetuser',
    u'posixaccount',
    u'krbprincipalaux',
    u'krbticketpolicyaux',
    u'ipaobject',
    u'ipasshuser',
    u'ipaSshGroupOfPubKeys',
]

user = user_base + [u'mepOriginEntry']

group = [
    u'top',
    u'groupofnames',
    u'nestedgroup',
    u'ipausergroup',
    u'ipaobject',
]

externalgroup = group + [u'ipaexternalgroup']
posixgroup = group + [u'posixgroup']

host = [
    u'ipasshhost',
    u'ipaSshGroupOfPubKeys',
    u'ieee802device',
    u'ipaobject',
    u'nshost',
    u'ipahost',
    u'pkiuser',
    u'ipaservice',
    u'krbprincipalaux',
    u'krbprincipal',
    u'top',
]

hostgroup = [
    u'ipaobject',
    u'ipahostgroup',
    u'nestedGroup',
    u'groupOfNames',
    u'top',
    u'mepOriginEntry',
]

role = [
    u'groupofnames',
    u'nestedgroup',
    u'top',
]

system_permission = [
    u'groupofnames',
    u'ipapermission',
    u'top'
]

permission = system_permission + [
    u'ipapermissionv2',
]

privilege = [
    u'nestedgroup',
    u'groupofnames',
    u'top'
]

service = [
    u'krbprincipal',
    u'krbprincipalaux',
    u'krbticketpolicyaux',
    u'ipaobject',
    u'ipaservice',
    u'pkiuser',
    u'ipakrbprincipal',
    u'top',
]

hbacsvc = [
    u'ipaobject',
    u'ipahbacservice',
]

hbacsvcgroup = [
    u'ipaobject',
    u'ipahbacservicegroup',
    u'groupOfNames',
    u'top',
]

sudocmd = [
    u'ipaobject',
    u'ipasudocmd',
]

sudocmdgroup = [
    u'ipaobject',
    u'ipasudocmdgrp',
    u'groupOfNames',
    u'top',
]

netgroup = [
    u'ipaobject',
    u'ipaassociation',
    u'ipanisnetgroup',
]

automember = [
    u'top',
    u'automemberregexrule',
]

selinuxusermap = [
    u'ipaassociation',
    u'ipaselinuxusermap',
]

hbacrule = [
    u'ipaassociation',
    u'ipahbacrule',
]

dnszone = [
    u'top',
    u'idnsrecord',
    u'idnszone',
]

dnsforwardzone = [
    u'top',
    u'idnsforwardzone',
]

dnsrecord = [
    u'top',
    u'idnsrecord',
]

realmdomains = [
    u'top',
    u'nsContainer',
    u'domainRelatedObject',
]

radiusproxy = [
    u'ipatokenradiusconfiguration',
    u'top',
]

pwpolicy = [
    u'krbpwdpolicy',
    u'nscontainer',
    u'top',
]

idview = [
    u'ipaIDView',
    u'nsContainer',
    u'top'
]

idoverrideuser = [
    u'ipaOverrideAnchor',
    u'top',
    u'ipaUserOverride',
    u'ipasshuser',
    u'ipaSshGroupOfPubKeys'
]

idoverridegroup = [
    u'ipaOverrideAnchor',
    u'top',
    u'ipaGroupOverride',
]

servicedelegationrule = [
    u'top',
    u'groupofprincipals',
    u'ipakrb5delegationacl',
]

servicedelegationtarget = [
    u'top',
    u'groupofprincipals',
]

certprofile = [
    u'top',
    u'ipacertprofile',
]

caacl = [
    u'ipaassociation',
    u'ipacaacl'
]

ca = [
    u'top',
    u'ipaca',
]

certmaprule = [
    u'top',
    u'ipacertmaprule',
]

certmapconfig = [
    u'top',
    u'nsContainer',
    u'ipaCertMapConfigObject',
]



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

from ..compat import CompatCommand, CompatMethod, CompatObject

Object = CompatObject


class Command(CompatCommand):
    api_version = u'2.114'


class Method(Command, CompatMethod):
    pass



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

ENTITY = 'idview'
USER_FACET = 'idoverrideuser'
GROUP_FACET = 'idoverridegroup'
HOST_FACET = 'appliedtohosts'

PKEY = 'itest-view'
DATA = {
    'pkey': PKEY,
    'add': [
        ('textbox', 'cn', PKEY),
        ('textarea', 'description', 'Description of ID view'),
    ],
    'mod': [
        ('textarea', 'description', 'Different description'),
    ],
}



# Authors:
#   Petr Vobornik <pvoborni@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Automember tests
"""

from ipatests.test_webui.ui_driver import UI_driver, screenshot
import ipatests.test_webui.data_hostgroup as hostgroup
from ipatests.test_webui.test_host import host_tasks
import pytest

ENTITY = 'automember'

USER_GROUP_PKEY = 'admins'
USER_GROUP_DATA = {
    'pkey': USER_GROUP_PKEY,
    'add': [
        ('combobox', 'cn', USER_GROUP_PKEY),
    ],
    'mod': [
        ('textarea', 'description', 'user group rule description'),
    ],
}

HOST_GROUP_DATA = {
    'pkey': hostgroup.PKEY,
    'add': [
        ('combobox', 'cn', hostgroup.PKEY),
    ],
    'mod': [
        ('textarea', 'description', 'host group rule description'),
    ],
}

SEARCH_CASES = {
    'name': 'search-123',
    'description': 'Short description !@#$%^&*()',
    'positive': [
        'search-123',
        'search',
        'search ',
        ' search',
        'SEARCH',
        '123',
        '!@#$%^&*()',
        'hort descr',
        'description  !',
    ],
    'negative': [
        'searc-123',
        '321',
        'search 123',
        'search Short',
        'description!',
    ],
}

# Condition types
INCLUSIVE = 'inclusive'
EXCLUSIVE = 'exclusive'


@pytest.mark.tier1
class TestAutomember(UI_driver):

    AUTOMEMBER_RULE_EXISTS_ERROR = (
        'Automember rule with name "{}" already exists'
    )

    def setup(self):
        super(TestAutomember, self).setup()
        self.init_app()

    def add_user_group_rules(self, *pkeys, **kwargs):
        # We implicitly trigger "Add and Add Another" by passing multiple
        # records to add_record method.
        # TODO: Create more transparent mechanism to test "Add <entity>" dialog
        self.add_record(
            ENTITY,
            [{
                'pkey': pkey,
                'add': [('combobox', 'cn', pkey)],
            } for pkey in pkeys],
            facet='searchgroup',
            **kwargs
        )

    def add_host_group_rules(self, *pkeys, **kwargs):
        self.add_record(
            ENTITY,
            [{
                'pkey': pkey,
                'add': [('combobox', 'cn', pkey)],
            } for pkey in pkeys],
            facet='searchhostgroup',
            **kwargs
        )

    def add_user(self, pkey, name, surname):
        self.add_record('user', {
            'pkey': pkey,
            'add': [
                ('textbox', 'uid', pkey),
                ('textbox', 'givenname', name),
                ('textbox', 'sn', surname),
            ]
        })

    def add_user_group(self, pkey, description=''):
        self.add_record('group', {
            'pkey': pkey,
            'add': [
                ('textbox', 'cn', pkey),
                ('textarea', 'description', description),
            ]
        })

    def add_host_group(self, pkey, description=''):
        self.add_record('hostgroup', {
            'pkey': pkey,
            'add': [
                ('textbox', 'cn', pkey),
                ('textarea', 'description', description),
            ]
        })

    def delete_users(self, *pkeys):
        self.delete('user', [{'pkey': pkey} for pkey in pkeys])

    def delete_user_groups(self, *pkeys):
        self.delete('group', [{'pkey': pkey} for pkey in pkeys])

    def delete_user_group_rules(self, *pkeys):
        self.delete(ENTITY, [{'pkey': pkey} for pkey in pkeys],
                    facet='searchgroup')

    def delete_host_groups(self, *pkeys):
        self.delete('hostgroup', [{'pkey': pkey} for pkey in pkeys])

    def delete_host_group_rules(self, *pkeys):
        self.delete(ENTITY, [{'pkey': pkey} for pkey in pkeys],
                    facet='searchhostgroup')

    def add_conditions(self, conditions, condition_type):
        """
        Add conditions to a rule

        :param conditions: list of conditions where condition is a pair
                           (attribute, expression)
        :param condition_type: can be 'inclusive' or 'exclusive'
        """

        name = 'automember{}regex'.format(condition_type)

        attribute, expression = conditions[0]
        another_conditions = conditions[1:]
        another_conditions.reverse()

        self.add_table_record(name, {'fields': [
            ('selectbox', 'key', attribute),
            ('textbox', name, expression)
        ]}, add_another=bool(another_conditions))

        while another_conditions:
            attribute, expression = another_conditions.pop()
            self.add_another_table_record(
                {'fields': [
                    ('selectbox', 'key', attribute),
                    ('textbox', name, expression)
                ]},
                add_another=bool(another_conditions)
            )

    def delete_conditions(self, conditions, condition_type):
        """
        Delete rule conditions

        :param conditions: list of conditions where condition is a pair
                           (attribute, expression)
        :param condition_type: can be 'inclusive' or 'exclusive'
        """

        self.delete_record(
            ['{}={}'.format(attr, exp) for attr, exp in conditions],
            parent=self.get_facet(),
            table_name='automember{}regex'.format(condition_type)
        )

    def open_new_condition_dialog(self, condition_type):
        table = self.find_by_selector(
            "table[name='automember{}regex'].table".format(condition_type),
            strict=True
        )
        btn = self.find_by_selector(".btn[name=add]", table, strict=True)
        btn.click()
        self.wait()

    def get_host_util(self):
        host_util = host_tasks()
        host_util.driver = self.driver
        host_util.config = self.config
        return host_util

    @screenshot
    def test_crud(self):
        """
        Basic CRUD: automember
        """

        # user group rule
        self.basic_crud(ENTITY, USER_GROUP_DATA,
                        search_facet='searchgroup',
                        default_facet='usergrouprule',
                        details_facet='usergrouprule',
                        )

        # prepare host group
        self.basic_crud(hostgroup.ENTITY, hostgroup.DATA,
                        default_facet=hostgroup.DEFAULT_FACET,
                        delete=False)

        # host group rule
        self.navigate_by_menu('identity/automember/amhostgroup')

        self.basic_crud(ENTITY, HOST_GROUP_DATA,
                        search_facet='searchhostgroup',
                        default_facet='hostgrouprule',
                        details_facet='hostgrouprule',
                        navigate=False,
                        breadcrumb='Host group rules',
                        )

        # cleanup
        self.delete(hostgroup.ENTITY, [hostgroup.DATA])

    @screenshot
    def test_rebuild_membership_hosts(self):
        """
        Test automember rebuild membership feature for hosts
        """

        host_util = self.get_host_util()
        domain = self.config.get('ipa_domain')
        host1 = 'web1.%s' % domain
        host2 = 'web2.%s' % domain

        # Add a hostgroup
        self.add_host_group('webservers', 'webservers')

        # Add hosts
        self.add_record('host', host_util.get_data("web1", domain))
        self.add_record('host', host_util.get_data("web2", domain))

        # Add an automember rule
        self.add_host_group_rules('webservers')

        # Add a condition for automember rule
        self.navigate_to_record('webservers')
        self.add_conditions([('fqdn', '^web[1-9]+')], condition_type=INCLUSIVE)

        # Assert that hosts are not members of hostgroup
        self.navigate_to_record('webservers', entity='hostgroup')
        self.facet_button_click('refresh')
        self.wait_for_request()
        self.assert_record(host1, negative=True)
        self.assert_record(host2, negative=True)

        # Rebuild membership for first host, using action on host details facet
        self.navigate_to_record(host1, entity='host')
        self.action_list_action('automember_rebuild')

        # Assert that host is now a member of hostgroup
        self.navigate_to_record('webservers', entity='hostgroup')
        self.facet_button_click('refresh')
        self.wait_for_request()
        self.assert_record(host1)
        self.assert_record(host2, negative=True)

        # Remove host from hostgroup
        self.delete_record(host1)

        # Assert that host was re-added to the group.
        # The behavior is expected with the plugin default setting:
        # the entry cn=Auto Membership Plugin,cn=plugins,cn=config has
        # a default value autoMemberProcessModifyOps: on
        #
        # See https://www.port389.org/docs/389ds/design/automember-postop-modify-design.html # noqa: E501
        self.facet_button_click('refresh')
        self.wait_for_request()
        self.assert_record(host1)

        # Rebuild membership for all hosts, using action on hosts search facet
        self.navigate_by_menu('identity/host')
        self.action_list_action('automember_rebuild')

        # Assert that hosts are now members of hostgroup
        self.navigate_to_record('webservers', entity='hostgroup')
        self.facet_button_click('refresh')
        self.wait_for_request()
        self.assert_record(host1)
        self.assert_record(host2)

        # Delete hostgroup, hosts and automember rule
        self.delete('host', [{'pkey': host1}, {'pkey': host2}])
        self.delete_host_group_rules('webservers')
        self.delete_host_groups('webservers')

    @screenshot
    def test_rebuild_membership_users(self):
        """
        Test automember rebuild membership feature for users
        """

        # Add a group
        self.add_user_group('devel', 'devel')

        # Add users
        self.add_user('dev1', 'Dev', 'One')
        self.add_user('dev2', 'Dev', 'Two')

        # Add an automember rule
        self.add_user_group_rules('devel')

        # Add a condition for automember rule
        self.navigate_to_record('devel')
        self.add_conditions([('uid', '^dev[1-9]+')], condition_type=INCLUSIVE)

        # Assert that users are not members of group
        self.navigate_to_record('devel', entity='group')
        self.facet_button_click('refresh')
        self.wait_for_request()
        self.assert_record('dev1', negative=True)
        self.assert_record('dev2', negative=True)

        # Rebuild membership for first user, using action on user details facet
        self.navigate_to_record('dev1', entity='user')
        self.action_list_action('automember_rebuild')

        # Assert that user is now a member of group
        self.navigate_to_record('devel', entity='group')
        self.facet_button_click('refresh')
        self.wait_for_request()
        self.assert_record('dev1')
        self.assert_record('dev2', negative=True)

        # Remove user from group
        self.delete_record('dev1')

        # Assert that user was re-added to the group
        # The behavior is expected with the plugin default setting:
        # the entry cn=Auto Membership Plugin,cn=plugins,cn=config has
        # a default value autoMemberProcessModifyOps: on
        #
        # See https://www.port389.org/docs/389ds/design/automember-postop-modify-design.html # noqa: E501
        self.facet_button_click('refresh')
        self.wait_for_request()
        self.assert_record('dev1')

        # Rebuild membership for all users, using action on users search facet
        self.navigate_by_menu('identity/user_search')
        self.action_list_action('automember_rebuild')

        # Assert that users are now members of group
        self.navigate_to_record('devel', entity='group')
        self.facet_button_click('refresh')
        self.wait_for_request()
        self.assert_record('dev1')
        self.assert_record('dev2')

        # Delete group, users and automember rule
        self.delete_users('dev1', 'dev2')
        self.delete_user_group_rules('devel')
        self.delete_user_groups('devel')

    @screenshot
    def test_add_multiple_user_group_rules(self):
        """
        Test creating and deleting multiple user group rules
        """

        groups = ['group1', 'group2', 'group3']

        for group in groups:
            self.add_user_group(group)

        self.add_user_group_rules(*groups)
        self.delete_user_group_rules(*groups)

    @screenshot
    def test_add_multiple_host_group_rules(self):
        """
        Test creating and deleting multiple host group rules
        """

        groups = ['group1', 'group2', 'group3']

        for group in groups:
            self.add_host_group(group)

        self.add_host_group_rules(*groups)
        self.delete_host_group_rules(*groups)

    @screenshot
    def test_search_user_group_rule(self):
        """
        Test searching user group rules using filter
        """

        pkey = SEARCH_CASES['name']
        self.add_user_group(pkey, '')
        self.add_user_group_rules(pkey)
        self.navigate_to_record(pkey)
        self.mod_record(ENTITY, {'mod': [
            ('textarea', 'description', SEARCH_CASES['description']),
        ]}, facet='usergrouprule')
        self.navigate_to_entity(ENTITY, facet='searchgroup')

        for text in SEARCH_CASES['positive']:
            self.apply_search_filter(text)
            self.wait_for_request()
            self.assert_record(pkey)

        for text in SEARCH_CASES['negative']:
            self.apply_search_filter(text)
            self.wait_for_request()
            self.assert_record(pkey, negative=True)

        self.delete_user_group_rules(pkey)
        self.delete_user_groups(pkey)

    @screenshot
    def test_search_host_group_rule(self):
        """
        Test searching host group rules using filter
        """

        pkey = SEARCH_CASES['name']
        self.add_host_group(pkey, '')
        self.add_host_group_rules(pkey, navigate=True)
        self.navigate_to_record(pkey)
        self.mod_record(ENTITY, {'mod': [
            ('textarea', 'description', SEARCH_CASES['description']),
        ]}, facet='hostgrouprule')
        self.navigate_to_entity(ENTITY, facet='searchhostgroup')

        for text in SEARCH_CASES['positive']:
            self.apply_search_filter(text)
            self.wait_for_request()
            self.assert_record(pkey)

        for text in SEARCH_CASES['negative']:
            self.apply_search_filter(text)
            self.wait_for_request()
            self.assert_record(pkey, negative=True)

        self.delete_host_group_rules(pkey)
        self.delete_host_groups(pkey)

    @screenshot
    def test_add_user_group_rule_conditions(self):
        """
        Test creating and deleting user group rule conditions
        """

        pkey = 'devel'
        one_inc_condition = ('employeetype', '*engineer*')
        inc_conditions = [
            ('cn', 'inclusive-expression'),
            ('description', 'other-inclusive-expression'),
        ]
        one_exc_condition = ('employeetype', '*manager*')
        exc_conditions = [
            ('cn', 'exclusive-expression'),
            ('description', 'other-exclusive-expression'),
        ]

        self.add_user_group(pkey)
        self.add_user_group_rules(pkey)

        self.navigate_to_record(pkey)

        self.add_conditions([one_inc_condition], condition_type=INCLUSIVE)
        self.add_conditions(inc_conditions, condition_type=INCLUSIVE)
        self.add_conditions([one_exc_condition], condition_type=EXCLUSIVE)
        self.add_conditions(exc_conditions, condition_type=EXCLUSIVE)

        self.delete_conditions([one_inc_condition], condition_type=INCLUSIVE)
        self.delete_conditions(inc_conditions, condition_type=INCLUSIVE)
        self.delete_conditions([one_exc_condition], condition_type=EXCLUSIVE)
        self.delete_conditions(inc_conditions, condition_type=EXCLUSIVE)

        self.delete_user_group_rules(pkey)
        self.delete_user_groups(pkey)

    @screenshot
    def test_add_host_group_rule_conditions(self):
        """
        Test creating and deleting user group rule conditions
        """

        pkey = 'devel'
        one_inc_condition = ('ipaclientversion', '4.8')
        inc_conditions = [
            ('cn', 'inclusive-expression'),
            ('description', 'other-inclusive-expression'),
        ]
        one_exc_condition = ('ipaclientversion', '4.7')
        exc_conditions = [
            ('cn', 'exclusive-expression'),
            ('description', 'other-exclusive-expression'),
        ]

        self.add_host_group(pkey)
        self.add_host_group_rules(pkey)

        self.navigate_to_record(pkey)

        self.add_conditions([one_inc_condition], condition_type=INCLUSIVE)
        self.add_conditions(inc_conditions, condition_type=INCLUSIVE)
        self.add_conditions([one_exc_condition], condition_type=EXCLUSIVE)
        self.add_conditions(exc_conditions, condition_type=EXCLUSIVE)

        self.delete_conditions([one_inc_condition], condition_type=INCLUSIVE)
        self.delete_conditions(inc_conditions, condition_type=INCLUSIVE)
        self.delete_conditions([one_exc_condition], condition_type=EXCLUSIVE)
        self.delete_conditions(inc_conditions, condition_type=EXCLUSIVE)

        self.delete_host_group_rules(pkey)
        self.delete_host_groups(pkey)

    @screenshot
    def test_cancel_new_user_group_rule_condition_dialog(self):
        """
        Test canceling of creating new user group rule condition
        """

        pkey = 'devel'

        self.add_user_group(pkey)
        self.add_user_group_rules(pkey)
        self.navigate_to_record(pkey)

        for condition_type in [INCLUSIVE, EXCLUSIVE]:
            self.open_new_condition_dialog(condition_type)
            self.fill_fields([('selectbox', 'key', 'title')])
            self.dialog_button_click('cancel')

        self.delete_user_group_rules(pkey)
        self.delete_user_groups(pkey)

    @screenshot
    def test_cancel_new_host_group_rule_condition_dialog(self):
        """
        Test canceling of creating new host group rule condition
        """

        pkey = 'devel'

        self.add_host_group(pkey)
        self.add_host_group_rules(pkey)
        self.navigate_to_record(pkey)

        for condition_type in [INCLUSIVE, EXCLUSIVE]:
            self.open_new_condition_dialog(condition_type)
            self.fill_fields([('selectbox', 'key', 'serverhostname')])
            self.dialog_button_click('cancel')

        self.delete_host_group_rules(pkey)
        self.delete_host_groups(pkey)

    @screenshot
    def test_set_default_user_group(self):
        """
        Test setting default user group
        """

        pkey = 'default-user-group'
        user_pkey = 'some-user'

        self.add_user_group(pkey)
        self.navigate_by_menu('identity/automember/amgroup')
        self.select_combobox('automemberdefaultgroup', pkey)

        self.add_user(user_pkey, 'Some', 'User')
        self.navigate_to_record(user_pkey)
        self.switch_to_facet('memberof_group')
        self.assert_record(pkey)

        # Clear default user group
        self.navigate_by_menu('identity/automember/amgroup')
        self.select_combobox('automemberdefaultgroup', '')

        self.delete_users(user_pkey)
        self.delete_user_groups(pkey)

    @screenshot
    def test_set_default_host_group(self):
        """
        Test setting default host group
        """

        pkey = 'default-host-group'
        host_util = self.get_host_util()
        domain = self.config.get('ipa_domain')

        self.add_host_group(pkey)
        self.navigate_by_menu('identity/automember/amhostgroup')
        self.select_combobox('automemberdefaultgroup', pkey)

        host_data = host_util.get_data('some-host', domain)
        self.add_record('host', host_data)
        self.navigate_to_record(host_data['pkey'])
        self.switch_to_facet('memberof_hostgroup')
        self.assert_record(pkey)

        # Clear default host group
        self.navigate_by_menu('identity/automember/amhostgroup')
        self.select_combobox('automemberdefaultgroup', '')

        self.delete('host', [{'pkey': host_data['pkey']}])
        self.delete_host_groups(pkey)

    @screenshot
    def test_add_user_group_rule_with_no_group(self):

        self.add_record(
            ENTITY,
            {'pkey': 'empty-user-group', 'add': []},
            facet='searchgroup',
            negative=True
        )

    @screenshot
    def test_add_host_group_rule_with_no_group(self):

        self.add_record(
            ENTITY,
            {'pkey': 'empty-host-group', 'add': []},
            facet='searchhostgroup',
            negative=True
        )

    @screenshot
    def test_add_user_group_rules_for_same_group(self):
        """
        Test creating user group rules for same group
        """

        group_name = 'some-user-group'

        self.add_user_group(group_name)
        self.add_user_group_rules(group_name)
        self.add_user_group_rules(group_name, negative=True, pre_delete=False)

        self.assert_last_error_dialog(
            self.AUTOMEMBER_RULE_EXISTS_ERROR.format(group_name)
        )

        self.delete_user_group_rules(group_name)
        self.delete_user_groups(group_name)

    @screenshot
    def test_add_host_group_rules_for_same_group(self):
        """
        Test creating host group rules for same group
        """

        group_name = 'some-host-group'

        self.add_host_group(group_name)
        self.add_host_group_rules(group_name)
        self.add_host_group_rules(group_name, negative=True, pre_delete=False)

        self.assert_last_error_dialog(
            self.AUTOMEMBER_RULE_EXISTS_ERROR.format(group_name)
        )

        self.delete_host_group_rules(group_name)
        self.delete_host_groups(group_name)

    @screenshot
    def test_cancel_group_rule_creating(self):
        """
        Test canceling of creating new automember group rule
        """

        self.add_user_group_rules('some-user-group', dialog_btn='cancel')
        self.add_host_group_rules('some-host-group', dialog_btn='cancel')



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#

"""
Test LoginScreen widget and all it's views
"""
import urllib

from ipatests.test_webui.ui_driver import UI_driver
from ipatests.test_webui.ui_driver import screenshot
import ipatests.test_webui.data_loginscreen as loginscreen

try:
    from selenium.webdriver.common.by import By
    from selenium.webdriver.common.keys import Keys
    from selenium.webdriver.support.wait import WebDriverWait
except ImportError:
    pass

import pytest


@pytest.mark.tier1
class TestLoginScreen(UI_driver):

    def setup(self, *args, **kwargs):
        super(TestLoginScreen, self).setup(*args, **kwargs)
        self.init_app()
        self.add_test_user()
        self.logout()

    def teardown(self, *args, **kwargs):
        # log out first
        if (self.logged_in()):
            self.logout()
        else:
            self.load_url(self.get_base_url())
        # log in as administrator
        self.login()
        self.delete_test_user()
        super(TestLoginScreen, self).teardown(*args, **kwargs)

    def delete_test_user(self):
        """
        Delete user for tests
        """
        # User is not logged in
        assert self.logged_in()
        self.navigate_to_entity(loginscreen.ENTITY)
        self.delete_record(loginscreen.PKEY)

    def add_test_user(self):
        """
        Add user for tests
        """
        # User is not logged in
        assert self.logged_in()
        self.add_record(loginscreen.ENTITY, loginscreen.DATA_ITEST_USER,
                        navigate=False)

    def assert_notification(self, type='success', assert_text=None,
                            link_text=None, link_url=None):
        """
        Assert whether we have a notification of particular type
        """

        notification_type = 'div.validation-summary .alert-{}'.format(type)
        # wait for a half sec for notification to appear
        self.wait(0.5)
        is_present = self.find(notification_type, By.CSS_SELECTOR)
        # Notification not present
        assert is_present
        if assert_text:
            assert assert_text in is_present.text

        if link_text and link_url:
            link = self.find_xelement(".//a", is_present)
            # Text on link placed on validation widget
            assert link_text == link.text
            # URL of link placed on validation widget
            assert link_url == link.get_attribute('href')

    def find_xelement(self, expression, parent, strict=True):
        """
        Find element by xpath related to a given parent
        """
        return self.find(expression, By.XPATH, parent, many=False,
                         strict=strict)

    def find_xelements(self, expression, parent, strict=True):
        """
        Find elements by xpath related to the given parent
        """
        return self.find(expression, By.XPATH, parent, many=True,
                         strict=strict)

    def button_click_on_login_screen(self, name):
        """
        Find a button with the given name on LoginScreen widget and
        then click on
        """
        login_screen = self.get_login_screen()
        button = self.find_xelement(".//button[@name='{}']".format(name),
                                    login_screen)
        assert button.is_displayed()
        button.click()

    def get_input_field(self, name, parent):
        """
        Find a input field with the given name and parent
        """
        return self.find_xelement(
            ".//input[@name='{}']".format(name),
            parent
        )

    def load_url(self, url):
        """
        Navigate to Web page and wait for loading of all dependencies.
        """
        self.driver.get(url)
        runner = self
        WebDriverWait(self.driver, 10).until(
            lambda d: runner.files_loaded()
        )

    def relogin_with_new_password(self):
        """
        Log out and then log in using a new password.
        It is need to check a new password
        """
        if (self.logged_in()):
            self.logout()
        else:
            self.load_url(self.get_base_url())
        self.login(loginscreen.PKEY, loginscreen.PASSWD_ITEST_USER_NEW)
        # User is not logged in
        assert self.logged_in()

    def reset_password(self, username=None, current_password=None,
                       new_password=None, link_text=None, link_url=None):
        """
        Reset password with the given one
        """
        login_screen = self.get_login_screen()

        if username is not None:
            username_field = self.get_input_field('username', login_screen)
        cur_pass_field = self.get_input_field('current_password', login_screen)
        new_pass_field = self.get_input_field('new_password', login_screen)
        verify_pass_field = self.get_input_field('verify_password',
                                                 login_screen)
        if username is not None:
            username_field.send_keys(username)
        cur_pass_field.send_keys(current_password)
        new_pass_field.send_keys(new_password)
        verify_pass_field.send_keys(new_password)
        verify_pass_field.send_keys(Keys.RETURN)
        self.wait()
        self.assert_notification(assert_text='Password change complete',
                                 link_text=link_text, link_url=link_url)

    def get_data_from_form_row(self, form_row):
        """
        Parse data from the form record to a comparable structure
        """
        result = []
        label = self.find_xelement(".//label", form_row)
        assert label.is_displayed()
        result.append(label.get_attribute('name'))
        result.append(label.text)
        req = self.find_xelement("./../..", label)
        result.append(self.has_class(req, 'required'))

        field = self.find_xelement(".//input", form_row)
        # not editable field
        editable = field.is_displayed()
        if not editable:
            field = self.find_xelement(".//p", form_row)

        result.append(editable)
        assert field.is_displayed()
        result.append(field.get_attribute('type'))
        result.append(field.get_attribute('name'))
        if not editable:
            result.append(field.text)
        else:
            result.append(field.get_attribute('value'))
        result.append(field.get_attribute('placeholder'))

        return tuple(result)

    def get_data_from_button(self, button):
        """
        Parse data from the button to a comparable structure
        """
        result = []
        result.append(button.get_attribute('name'))
        result.append(button.get_attribute('title'))

        return tuple(result)

    def assert_form_equals(self, actual_form, expected_form):
        """
        Compare two forms
        """
        assert len(actual_form) == len(expected_form)
        for act_row, exp_row in zip(actual_form, expected_form):
            # structure of rows
            # label_name, label_text,
            # required, editable,
            # input_type, input_name,
            # input_text, placeholder
            assert self.get_data_from_form_row(act_row) == exp_row

    def assert_buttons_equal(self, actual_buttons, expected_buttons):
        """
        Compare button sets
        """
        assert len(actual_buttons) == len(expected_buttons)
        for act_button, exp_button in zip(actual_buttons, expected_buttons):
            assert self.get_data_from_button(act_button) == exp_button

    def assert_validations_equal(self, actual_alerts, expected_alerts):
        """
        Compare validation sets
        """
        assert len(actual_alerts) == len(expected_alerts)
        for act_alert, exp_alert in zip(actual_alerts, expected_alerts):
            assert (act_alert.text,) == exp_alert

    def has_validation(self, parent):
        return self.find_xelement(".//div[@name='validation']", parent,
                                  strict=False)

    def check_elements_of_form(self, form_data):

        login_screen = self.get_login_screen()
        form = self.find_xelement(".//div[@class='form-horizontal']",
                                  login_screen)
        # rows
        form_rows = self.find_xelements(
            ".//div[contains(@class, 'form-group')]", form
        )

        form_rows = [el for el in form_rows if el.is_displayed() and not
                     self.has_validation(el)]
        self.assert_form_equals(form_rows, form_data['rows'])

        # buttons
        buttons = self.find_xelements(".//button", login_screen)
        buttons = [el for el in buttons if el.is_displayed()]
        self.assert_buttons_equal(buttons, form_data['buttons'])

    def check_alerts(self, form_data):
        login_screen = self.get_login_screen()

        # Push the the most rigth button to see the Required fields
        # it should be either 'Reset' or 'Reset and Login' or 'Login' button
        self.button_click_on_login_screen(form_data['buttons'][-1][0])

        alerts = self.find_xelements(
            ".//*[@data-name][contains(@class, 'alert-danger')]", login_screen
        )
        required_msgs = form_data['required_msg']
        self.assert_validations_equal(alerts, required_msgs)

    def load_reset_and_login_view(self):

        self.load()
        assert self.login_screen_visible()
        username = loginscreen.PKEY
        current_password = loginscreen.PASSWD_ITEST_USER

        login_screen = self.get_login_screen()
        username_field = self.get_input_field('username', login_screen)
        cur_pass_field = self.get_input_field('password', login_screen)
        username_field.send_keys(username)
        cur_pass_field.send_keys(current_password)
        cur_pass_field.send_keys(Keys.RETURN)
        self.wait()
        self.assert_notification(
            type='info',
            assert_text=(
                'Your password has expired. Please enter a new password.'
            )
        )

    def check_cancel(self):
        """
        Check 'login' view after a cancel of password reset
        """
        self.button_click_on_login_screen('cancel')
        self.check_elements_of_form(loginscreen.FILLED_LOGIN_FORM)

    @screenshot
    def test_reset_password_view(self):

        self.load_url('/'.join((self.get_base_url(), 'reset_password.html')))
        assert self.login_screen_visible()

        self.check_elements_of_form(loginscreen.RESET_PASSWORD_FORM)
        self.check_alerts(loginscreen.RESET_PASSWORD_FORM)

        username = loginscreen.PKEY
        current_password = loginscreen.PASSWD_ITEST_USER
        new_password = loginscreen.PASSWD_ITEST_USER_NEW
        self.reset_password(username, current_password, new_password)
        self.relogin_with_new_password()

    @screenshot
    def test_reset_password_view_with_redirect(self):

        redir_url = self.get_base_url().lower()
        encoded_redir_url = urllib.parse.urlencode({'url': redir_url})
        target_url = '/'.join((self.get_base_url(), 'reset_password.html?{}'))
        self.load_url(target_url.format(encoded_redir_url))
        assert self.login_screen_visible()

        self.check_elements_of_form(loginscreen.RESET_PASSWORD_FORM)
        self.check_alerts(loginscreen.RESET_PASSWORD_FORM)

        username = loginscreen.PKEY
        current_password = loginscreen.PASSWD_ITEST_USER
        new_password = loginscreen.PASSWD_ITEST_USER_NEW
        self.reset_password(username, current_password, new_password,
                            link_text='Continue to next page',
                            link_url=redir_url,
                            )
        self.relogin_with_new_password()

    @screenshot
    def test_reset_password_view_with_delayed_redirect(self):

        redir_url = self.get_base_url().lower() + '/'
        encoded_redir_url = urllib.parse.urlencode(
            {'url': redir_url, 'delay': 5}
        )
        target_url = '/'.join((self.get_base_url(), 'reset_password.html?{}'))
        self.load_url(target_url.format(encoded_redir_url))
        assert self.login_screen_visible()

        self.check_elements_of_form(loginscreen.RESET_PASSWORD_FORM)
        self.check_alerts(loginscreen.RESET_PASSWORD_FORM)
        username = loginscreen.PKEY
        current_password = loginscreen.PASSWD_ITEST_USER
        new_password = loginscreen.PASSWD_ITEST_USER_NEW
        self.reset_password(username, current_password, new_password,
                            link_text='Continue to next page',
                            link_url=redir_url,
                            )
        self.assert_notification(type='info',
                                 assert_text='You will be redirected in ')
        self.wait(3)
        # check url after start delay timer, but before end
        assert self.driver.current_url != redir_url
        self.wait(5)
        assert self.driver.current_url == redir_url
        self.relogin_with_new_password()

    @screenshot
    def test_reset_password_and_login_view(self):

        self.load_reset_and_login_view()
        self.check_elements_of_form(loginscreen.RESET_AND_LOGIN_FORM)
        self.check_alerts(loginscreen.RESET_AND_LOGIN_FORM)

        # check click on 'Cancel' button
        self.check_cancel()
        self.button_click_on_login_screen('login')

        # check if user is not logged
        assert not self.logged_in()
        current_password = loginscreen.PASSWD_ITEST_USER
        new_password = loginscreen.PASSWD_ITEST_USER_NEW
        # username is already here, do not fill up
        self.reset_password(current_password=current_password,
                            new_password=new_password)
        # waiting for auto login process
        self.wait(5)
        assert self.logged_in()
        # check again if new password is valid
        self.relogin_with_new_password()

    @screenshot
    def test_login_view(self):

        self.load()

        # check empty 'login' view
        self.check_elements_of_form(loginscreen.EMPTY_LOGIN_FORM)
        self.check_alerts(loginscreen.EMPTY_LOGIN_FORM)

        # check non empty 'login' view
        login_screen = self.get_login_screen()
        username_field = self.get_input_field('username', login_screen)
        username_field.send_keys(loginscreen.PKEY)
        self.wait(0.5)
        self.check_elements_of_form(loginscreen.LOGIN_FORM)
        self.check_alerts(loginscreen.LOGIN_FORM)



#
# Authors:
#   Rob Crittenden <rcritten@redhat.com>
#
# Copyright (C) 2011  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Provide a custom certificate used in the service tests.

The certificate in cached in a global variable so it only has to be created
once per test run.
"""

from __future__ import absolute_import

import os
import tempfile
import shutil
import base64
import re

from ipalib import api, x509
from ipaserver.plugins import rabase
from ipapython import certdb
from ipapython.dn import DN
from ipaplatform.paths import paths

_subject_base = None


def subject_base():
    global _subject_base

    if _subject_base is None:
        config = api.Command['config_show']()['result']
        _subject_base = DN(config['ipacertificatesubjectbase'][0])

    return _subject_base


def strip_cert_header(pem):
    """
    Remove the header and footer from a certificate.
    """
    regexp = (
        r"^-----BEGIN CERTIFICATE-----(.*?)-----END CERTIFICATE-----"
    )
    s = re.search(regexp, pem, re.MULTILINE | re.DOTALL)
    if s is not None:
        return s.group(1)
    else:
        return pem


def get_testcert(subject, principal):
    """Get the certificate, creating it if it doesn't exist"""
    reqdir = tempfile.mkdtemp(prefix="tmp-")
    try:
        _testcert = makecert(reqdir, subject,
                             principal)
    finally:
        shutil.rmtree(reqdir)
    return strip_cert_header(_testcert.decode('utf-8'))


def makecert(reqdir, subject, principal):
    """
    Generate a certificate that can be used during unit testing.
    """

    ra = rabase.rabase(api)
    if (not os.path.exists(ra.client_certfile) and
            api.env.xmlrpc_uri == 'http://localhost:8888/ipa/xml'):
        raise AssertionError('The self-signed CA is not configured, '
                             'see ipatests/test_xmlrpc/test_cert.py')

    nssdb = certdb.NSSDatabase(nssdir=reqdir)
    with open(nssdb.pwd_file, "w") as f:
        # Create an empty password file
        f.write("\n")
    # create db
    nssdb.create_db()
    # create CSR
    csr_file = os.path.join(reqdir, 'req')
    nssdb.run_certutil([
        "-R", "-s", str(subject),
        "-o", csr_file,
        "-z", paths.GROUP,
        "-a"
    ])
    with open(csr_file, "rb") as f:
        csr = f.read().decode('ascii')

    res = api.Command['cert_request'](csr, principal=principal, add=True)
    cert = x509.load_der_x509_certificate(
        base64.b64decode(res['result']['certificate']))
    return cert.public_bytes(x509.Encoding.PEM)



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

"""
The framework core.
"""

import abc
import collections
import functools
import itertools
import sys

import six

from . import util
from .util import from_

__all__ = ['InvalidStateError', 'KnobValueError', 'Property', 'knob',
           'Configurable', 'group', 'Component', 'Composite']

NoneType = type(None)
builtin_type = type

# Configurable states
_VALIDATE_PENDING = 'VALIDATE_PENDING'
_VALIDATE_RUNNING = 'VALIDATE_RUNNING'
_EXECUTE_PENDING = 'EXECUTE_PENDING'
_EXECUTE_RUNNING = 'EXECUTE_RUNNING'
_STOPPED = 'STOPPED'
_FAILED = 'FAILED'
_CLOSED = 'CLOSED'

_missing = object()
_counter = itertools.count()


@functools.cmp_to_key
def _class_key(a, b):
    if a is b:
        return 0
    elif issubclass(a, b):
        return -1
    elif issubclass(b, a):
        return 1
    else:
        return 0


class InvalidStateError(Exception):
    pass


class KnobValueError(ValueError):
    def __init__(self, name, message):
        super(KnobValueError, self).__init__(message)
        self.name = name


class PropertyBase(six.with_metaclass(util.InnerClassMeta, object)):
    # shut up pylint
    __outer_class__ = None
    __outer_name__ = None

    _order = None

    @property
    def default(self):
        raise AttributeError('default')

    def __init__(self, outer):
        pass

    def __get__(self, obj, obj_type):
        while obj is not None:
            try:
                return obj.__dict__[self.__outer_name__]
            except KeyError:
                pass
            obj = obj._get_fallback()

        try:
            return self.default
        except AttributeError:
            raise AttributeError(self.__outer_name__)

    def __set__(self, obj, value):
        try:
            obj.__dict__[self.__outer_name__] = value
        except KeyError:
            raise AttributeError(self.__outer_name__)

    def __delete__(self, obj):
        try:
            del obj.__dict__[self.__outer_name__]
        except KeyError:
            raise AttributeError(self.__outer_name__)


def Property(default=_missing):
    class_dict = {}
    if default is not _missing:
        class_dict['default'] = default

    return util.InnerClassMeta('Property', (PropertyBase,), class_dict)


class KnobBase(PropertyBase):
    type = None
    sensitive = False
    deprecated = False
    description = None
    cli_names = (None,)
    cli_deprecated_names = ()
    cli_metavar = None

    def __init__(self, outer):
        self.outer = outer

    def validate(self, value):
        pass

    @classmethod
    def group(cls):
        return cls.__outer_class__.group()

    @classmethod
    def is_cli_positional(cls):
        return all(n is not None and not n.startswith('-')
                   for n in cls.cli_names)

    @classmethod
    def default_getter(cls, func):
        @property
        def default(self):
            return func(self.outer)
        cls.default = default

        return cls

    @classmethod
    def validator(cls, func):
        def validate(self, value):
            func(self.outer, value)
            super(cls, self).validate(value)
        cls.validate = validate

        return cls


def _knob(type=_missing, default=_missing, bases=_missing, _order=_missing,
          sensitive=_missing, deprecated=_missing, description=_missing,
          group=_missing, cli_names=_missing, cli_deprecated_names=_missing,
          cli_metavar=_missing):
    if type is None:
        type = NoneType

    if bases is _missing:
        bases = (KnobBase,)
    elif isinstance(bases, builtin_type):
        bases = (bases,)

    if cli_names is None or isinstance(cli_names, str):
        cli_names = (cli_names,)
    elif cli_names is not _missing:
        cli_names = tuple(cli_names)

    if isinstance(cli_deprecated_names, str):
        cli_deprecated_names = (cli_deprecated_names,)
    elif cli_deprecated_names is not _missing:
        cli_deprecated_names = tuple(cli_deprecated_names)

    class_dict = {}
    if type is not _missing:
        class_dict['type'] = type
    if default is not _missing:
        class_dict['default'] = default
    if _order is not _missing:
        class_dict['_order'] = _order
    if sensitive is not _missing:
        class_dict['sensitive'] = sensitive
    if deprecated is not _missing:
        class_dict['deprecated'] = deprecated
    if description is not _missing:
        class_dict['description'] = description
    if group is not _missing:
        class_dict['group'] = group
    if cli_names is not _missing:
        class_dict['cli_names'] = cli_names
    if cli_deprecated_names is not _missing:
        class_dict['cli_deprecated_names'] = cli_deprecated_names
    if cli_metavar is not _missing:
        class_dict['cli_metavar'] = cli_metavar

    return util.InnerClassMeta('Knob', bases, class_dict)


def knob(type, default=_missing, **kwargs):
    """
    Define a new knob.
    """
    return _knob(
        type, default,
        _order=next(_counter),
        **kwargs
    )


def extend_knob(base, default=_missing, bases=_missing, group=_missing,
                **kwargs):
    """
    Extend an existing knob.
    """
    if bases is _missing:
        bases = (base,)

    if group is _missing:
        group = staticmethod(base.group)

    return _knob(
        _missing, default,
        bases=bases,
        _order=_missing,
        group=group,
        **kwargs
    )


class Configurable(six.with_metaclass(abc.ABCMeta, object)):
    """
    Base class of all configurables.

    FIXME: details of validate/execute, properties and knobs
    """

    @classmethod
    def properties(cls):
        """
        Iterate over properties defined for the configurable.
        """

        assert not hasattr(super(Configurable, cls), 'properties')

        seen = set()

        for owner_cls in cls.__mro__:
            result = []

            for name, prop_cls in owner_cls.__dict__.items():
                if name in seen:
                    continue
                seen.add(name)

                if not isinstance(prop_cls, type):
                    continue
                if not issubclass(prop_cls, PropertyBase):
                    continue

                result.append((prop_cls._order, owner_cls, name))

            result = sorted(result, key=lambda r: r[0])

            for _order, owner_cls, name in result:
                yield owner_cls, name

    @classmethod
    def knobs(cls):
        for owner_cls, name in cls.properties():
            prop_cls = getattr(owner_cls, name)
            if issubclass(prop_cls, KnobBase):
                yield owner_cls, name

    @classmethod
    def group(cls):
        assert not hasattr(super(Configurable, cls), 'group')

    def __init__(self, **kwargs):
        """
        Initialize the configurable.
        """

        cls = self.__class__
        for owner_cls, name in cls.properties():
            if name.startswith('_'):
                continue
            prop_cls = getattr(owner_cls, name)
            if not isinstance(prop_cls, type):
                continue
            if not issubclass(prop_cls, PropertyBase):
                continue

            try:
                value = kwargs.pop(name)
            except KeyError:
                pass
            else:
                setattr(self, name, value)

        for owner_cls, name in cls.knobs():
            if name.startswith('_'):
                continue
            if not isinstance(self, owner_cls):
                continue
            value = getattr(self, name, None)
            if value is None:
                continue

            prop_cls = getattr(owner_cls, name)
            prop = prop_cls(self)
            try:
                prop.validate(value)
            except ValueError as e:
                raise KnobValueError(name, str(e))

        if kwargs:
            extra = sorted(kwargs)
            raise TypeError(
                "{0}() got {1} unexpected keyword arguments: {2}".format(
                    type(self).__name__,
                    len(extra),
                    ', '.join(repr(name) for name in extra)))

        self._reset()

    def _reset(self):
        assert not hasattr(super(Configurable, self), '_reset')

        self.__state = _VALIDATE_PENDING
        self.__gen = util.run_generator_with_yield_from(self._configure())

    def _get_components(self):
        assert not hasattr(super(Configurable, self), '_get_components')

        raise TypeError("{0} is not composite".format(self))

    def _get_fallback(self):
        pass

    @abc.abstractmethod
    def _configure(self):
        """
        Coroutine which defines the logic of the configurable.
        """

        assert not hasattr(super(Configurable, self), '_configure')

        self.__transition(_VALIDATE_RUNNING, _EXECUTE_PENDING)

        while self.__state != _EXECUTE_RUNNING:
            yield

    def run(self):
        """
        Run the configurable.
        """

        self.validate()
        if self.__state == _EXECUTE_PENDING:
            return self.execute()
        return None

    def validate(self):
        """
        Run the validation part of the configurable.
        """

        for _nothing in self._validator():
            pass

    def _validator(self):
        """
        Coroutine which runs the validation part of the configurable.
        """

        return self.__runner(_VALIDATE_PENDING,
                             _VALIDATE_RUNNING,
                             self._handle_validate_exception)

    def execute(self):
        """
        Run the execution part of the configurable.
        """
        return_value = 0

        for rval in self._executor():
            if rval is not None and rval > return_value:
                return_value = rval

        return return_value

    def _executor(self):
        """
        Coroutine which runs the execution part of the configurable.
        """

        return self.__runner(_EXECUTE_PENDING,
                             _EXECUTE_RUNNING,
                             self._handle_execute_exception)

    def done(self):
        """
        Return True if the configurable has finished.
        """

        return self.__state in (_STOPPED, _FAILED, _CLOSED)

    def run_until_executing(self, gen):
        while self.__state != _EXECUTE_RUNNING:
            try:
                yield next(gen)
            except StopIteration:
                break

    def __runner(self, pending_state, running_state, exc_handler):
        self.__transition(pending_state, running_state)

        step = lambda: next(self.__gen)
        while True:
            try:
                step()
            except StopIteration:
                self.__transition(running_state, _STOPPED)
                break
            except GeneratorExit:
                self.__transition(running_state, _CLOSED)
                break
            except BaseException:
                exc_info = sys.exc_info()
                try:
                    exc_handler(exc_info)
                except BaseException:
                    self.__transition(running_state, _FAILED)
                    raise

            if self.__state != running_state:
                break

            try:
                yield
            except BaseException:
                exc_info = sys.exc_info()
                step = lambda: self.__gen.throw(*exc_info)
            else:
                step = lambda: next(self.__gen)

    def _handle_exception(self, exc_info):
        assert not hasattr(super(Configurable, self), '_handle_exception')

        six.reraise(*exc_info)

    def _handle_validate_exception(self, exc_info):
        assert not hasattr(super(Configurable, self),
                           '_handle_validate_exception')
        self._handle_exception(exc_info)

    def _handle_execute_exception(self, exc_info):
        assert not hasattr(super(Configurable, self),
                           '_handle_execute_exception')
        self._handle_exception(exc_info)

    def __transition(self, from_state, to_state):
        if self.__state != from_state:
            raise InvalidStateError(self.__state)

        self.__state = to_state


def group(cls):
    def group():
        return cls

    cls.group = staticmethod(group)

    return cls


class ComponentMeta(util.InnerClassMeta, abc.ABCMeta):
    pass


class ComponentBase(six.with_metaclass(ComponentMeta, Configurable)):
    # shut up pylint
    __outer_class__ = None
    __outer_name__ = None

    _order = None

    @classmethod
    def group(cls):
        result = super(ComponentBase, cls).group()
        if result is not None:
            return result
        else:
            return cls.__outer_class__.group()

    def __init__(self, parent, **kwargs):
        self.__parent = parent

        super(ComponentBase, self).__init__(**kwargs)

    @property
    def parent(self):
        return self.__parent

    def __get__(self, obj, obj_type):
        obj.__dict__[self.__outer_name__] = self
        return self

    def _get_fallback(self):
        return self.__parent

    def _handle_exception(self, exc_info):
        try:
            super(ComponentBase, self)._handle_exception(exc_info)
        except BaseException:
            exc_info = sys.exc_info()
            self.__parent._handle_exception(exc_info)


def Component(cls):
    class_dict = {}
    class_dict['_order'] = next(_counter)

    return ComponentMeta('Component', (ComponentBase, cls), class_dict)


class Composite(Configurable):
    """
    Configurable composed of any number of components.

    Provides knobs of all child components.
    """

    @classmethod
    def properties(cls):
        name_dict = {}
        owner_dict = collections.OrderedDict()

        for owner_cls, name in super(Composite, cls).properties():
            name_dict[name] = owner_cls
            owner_dict.setdefault(owner_cls, []).append(name)

        for owner_cls, name in cls.components():
            comp_cls = getattr(cls, name)

            for owner_cls, name in comp_cls.knobs():
                if hasattr(cls, name):
                    continue

                try:
                    last_owner_cls = name_dict[name]
                except KeyError:
                    name_dict[name] = owner_cls
                    owner_dict.setdefault(owner_cls, []).append(name)
                else:
                    knob_cls = getattr(owner_cls, name)
                    last_knob_cls = getattr(last_owner_cls, name)
                    if issubclass(knob_cls, last_knob_cls):
                        name_dict[name] = owner_cls
                        owner_dict[last_owner_cls].remove(name)
                        owner_dict.setdefault(owner_cls, [])
                        if name not in owner_dict[owner_cls]:
                            owner_dict[owner_cls].append(name)
                    elif not issubclass(last_knob_cls, knob_cls):
                        raise TypeError("{0}.knobs(): conflicting definitions "
                                        "of '{1}' in {2} and {3}".format(
                                            cls.__name__,
                                            name,
                                            last_owner_cls.__name__,
                                            owner_cls.__name__))

        for owner_cls in sorted(owner_dict, key=_class_key):
            for name in owner_dict[owner_cls]:
                yield owner_cls, name

    @classmethod
    def components(cls):
        assert not hasattr(super(Composite, cls), 'components')

        seen = set()

        for owner_cls in cls.__mro__:
            result = []

            for name, comp_cls in owner_cls.__dict__.items():
                if name in seen:
                    continue
                seen.add(name)

                if not isinstance(comp_cls, type):
                    continue
                if not issubclass(comp_cls, ComponentBase):
                    continue

                result.append((comp_cls._order, owner_cls, name))

            result = sorted(result, key=lambda r: r[0])

            for _order, owner_cls, name in result:
                yield owner_cls, name

    def __getattr__(self, name):
        for owner_cls, knob_name in self.knobs():
            if knob_name == name:
                break
        else:
            raise AttributeError(name)

        for component in self.__components:
            if isinstance(component, owner_cls):
                break
        else:
            raise AttributeError(name)

        return getattr(component, name)

    def _reset(self):
        self.__components = list(self._get_components())

        super(Composite, self)._reset()

    def _get_components(self):
        for _owner_cls, name in self.components():
            yield getattr(self, name)

    def _configure(self):
        validate = [(c, c._validator()) for c in self.__components]
        while True:
            new_validate = []
            for child, validator in validate:
                try:
                    next(validator)
                except StopIteration:
                    pass
                else:
                    new_validate.append((child, validator))
            if not new_validate:
                break
            validate = new_validate

            yield

        if not self.__components:
            return

        yield from_(super(Composite, self)._configure())

        execute = [(c, c._executor()) for c in self.__components
            if not c.done()]
        while True:
            new_execute = []
            for child, executor in execute:
                try:
                    next(executor)
                except StopIteration:
                    pass
                else:
                    new_execute.append((child, executor))
            if not new_execute:
                break
            execute = new_execute

            yield



# Authors: Karl MacMillan <kmacmill@redhat.com>
#
# Copyright (C) 2007  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import

# pylint: disable=deprecated-module
from optparse import (
    Option, Values, OptionParser, IndentedHelpFormatter, OptionValueError)
# pylint: enable=deprecated-module
from copy import copy
from configparser import SafeConfigParser
from urllib.parse import urlsplit
import socket
import functools

from dns.exception import DNSException
import dns.name

from ipaplatform.paths import paths
from ipapython.dn import DN
from ipapython.dnsutil import query_srv
from ipapython.ipautil import CheckedIPAddress, CheckedIPAddressLoopback


class IPAConfigError(Exception):
    def __init__(self, msg=''):
        self.msg = msg
        Exception.__init__(self, msg)

    def __repr__(self):
        return self.msg

    __str__ = __repr__

class IPAFormatter(IndentedHelpFormatter):
    """Our own optparse formatter that indents multiple lined usage string."""
    def format_usage(self, usage):
        usage_string = "Usage:"
        spacing = " " * len(usage_string)
        lines = usage.split("\n")
        ret = "%s %s\n" % (usage_string, lines[0])
        for line in lines[1:]:
            ret += "%s %s\n" % (spacing, line)
        return ret


def check_ip_option(option, opt, value, allow_loopback=False):
    try:
        if allow_loopback:
            return CheckedIPAddressLoopback(value)
        else:
            return CheckedIPAddress(value)
    except Exception as e:
        raise OptionValueError("option {}: invalid IP address {}: {}"
                               .format(opt, value, e))

def check_dn_option(option, opt, value):
    try:
        return DN(value)
    except Exception as e:
        raise OptionValueError("option %s: invalid DN: %s" % (opt, e))


def check_constructor(option, opt, value):
    con = option.constructor
    assert con is not None, "Oops! Developer forgot to set 'constructor' kwarg"
    try:
        return con(value)
    except Exception as e:
        raise OptionValueError("option {} invalid: {}".format(opt, e))


class IPAOption(Option):
    """
    optparse.Option subclass with support of options labeled as
    security-sensitive such as passwords.
    """
    ATTRS = Option.ATTRS + ["sensitive", "constructor"]
    TYPES = Option.TYPES + ("ip", "dn", "constructor", "ip_with_loopback")
    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
    TYPE_CHECKER["ip"] = check_ip_option
    TYPE_CHECKER["ip_with_loopback"] = functools.partial(check_ip_option,
                                                         allow_loopback=True)
    TYPE_CHECKER["dn"] = check_dn_option
    TYPE_CHECKER["constructor"] = check_constructor


class IPAOptionParser(OptionParser):
    """
    optparse.OptionParser subclass that uses IPAOption by default
    for storing options.
    """
    def __init__(self,
                 usage=None,
                 option_list=None,
                 option_class=IPAOption,
                 version=None,
                 conflict_handler="error",
                 description=None,
                 formatter=None,
                 add_help_option=True,
                 prog=None):
        OptionParser.__init__(self, usage, option_list, option_class,
                              version, conflict_handler, description,
                              formatter, add_help_option, prog)

    def get_safe_opts(self, opts):
        """
        Returns all options except those with sensitive=True in the same
        fashion as parse_args would
        """
        all_opts_dict = {
            o.dest: o for o in self._get_all_options()
            if hasattr(o, 'sensitive')
        }
        safe_opts_dict = {}

        for option, value in opts.__dict__.items():
            if all_opts_dict[option].sensitive != True:
                safe_opts_dict[option] = value

        return Values(safe_opts_dict)

def verify_args(parser, args, needed_args = None):
    """Verify that we have all positional arguments we need, if not, exit."""
    if needed_args:
        needed_list = needed_args.split(" ")
    else:
        needed_list = []
    len_need = len(needed_list)
    len_have = len(args)
    if len_have > len_need:
        parser.error("too many arguments")
    elif len_have < len_need:
        parser.error("no %s specified" % needed_list[len_have])


class IPAConfig:
    def __init__(self):
        self.default_realm = None
        self.default_server = []
        self.default_domain = None

    def get_realm(self):
        if self.default_realm:
            return self.default_realm
        else:
            raise IPAConfigError("no default realm")

    def get_server(self):
        if len(self.default_server):
            return self.default_server
        else:
            raise IPAConfigError("no default server")

    def get_domain(self):
        if self.default_domain:
            return self.default_domain
        else:
            raise IPAConfigError("no default domain")

# Global library config
config = IPAConfig()

def __parse_config(discover_server = True):
    p = SafeConfigParser()
    p.read(paths.IPA_DEFAULT_CONF)

    try:
        if not config.default_realm:
            config.default_realm = p.get("global", "realm")
    except Exception:
        pass
    if discover_server:
        try:
            s = p.get("global", "xmlrpc_uri")
            server = urlsplit(s)
            config.default_server.append(server.netloc)
        except Exception:
            pass
    try:
        if not config.default_domain:
            config.default_domain = p.get("global", "domain")
    except Exception:
        pass

def __discover_config(discover_server = True):
    servers = []
    try:
        if not config.default_domain:
            # try once with REALM -> domain
            domain = str(config.default_realm).lower()
            name = "_ldap._tcp." + domain

            try:
                servers = query_srv(name)
            except DNSException:
                # try cycling on domain components of FQDN
                try:
                    domain = dns.name.from_text(socket.getfqdn())
                except DNSException:
                    return False

                while True:
                    domain = domain.parent()

                    if str(domain) == '.':
                        return False
                    name = "_ldap._tcp.%s" % domain
                    try:
                        servers = query_srv(name)
                        break
                    except DNSException:
                        pass

            config.default_domain = str(domain).rstrip(".")

        if discover_server:
            if not servers:
                name = "_ldap._tcp.%s." % config.default_domain
                try:
                    servers = query_srv(name)
                except DNSException:
                    pass

            for server in servers:
                hostname = str(server.target).rstrip(".")
                config.default_server.append(hostname)

    except Exception:
        pass
    return None


def add_standard_options(parser):
    parser.add_option("--realm", dest="realm", help="Override default IPA realm")
    parser.add_option("--server", dest="server",
                      help="Override default FQDN of IPA server")
    parser.add_option("--domain", dest="domain", help="Override default IPA DNS domain")

def init_config(options=None):
    if options:
        config.default_realm = options.realm
        config.default_domain = options.domain
        if options.server:
            config.default_server.extend(options.server.split(","))

    if len(config.default_server):
        discover_server = False
    else:
        discover_server = True
    __parse_config(discover_server)
    __discover_config(discover_server)

    # make sure the server list only contains unique items
    new_server = []
    for server in config.default_server:
        if server not in new_server:
            new_server.append(server)
    config.default_server = new_server

    if not config.default_realm:
        raise IPAConfigError("IPA realm not found in DNS, in the config file (/etc/ipa/default.conf) or on the command line.")
    if not config.default_server:
        raise IPAConfigError("IPA server not found in DNS, in the config file (/etc/ipa/default.conf) or on the command line.")
    if not config.default_domain:
        raise IPAConfigError("IPA domain not found in the config file (/etc/ipa/default.conf) or on the command line.")
